import threading, copy, logging
from datetime import datetime
from ..singleton import Singleton
from ..models import FloatType, StringType, IntegerType, BooleanType
from ..modules.users.users import User
from ..modules.users.users import User
from ..utils.decorators import set_event, logging_error_handler
from ..filter import filter
# from ..iad import iad_outlier, iad_frozen_data, iad_out_of_range
from .tag import Tag
from flask_socketio import SocketIO

class CVT:
    """
    Current Value Table (CVT) class for a tag-based repository.

    This class is designed to hold in-memory tag-based values and manage observers for the required tags. It is intended to be used exclusively by PyAutomation and should not be used for other purposes.

    **Usage Example**:

    .. code-block:: python

        >>> from automation.tags import CVT
        >>> _cvt = CVT()

    """

    def __init__(self):
        
        self._tags = dict()
        self.data_types = ["float", "int", "bool", "str"]
        self.sio:SocketIO|None = None

    def set_socketio(self, sio:SocketIO):
        r"""
        Sets the SocketIO instance for real-time updates.

        **Parameters:**

        * **sio** (SocketIO): The SocketIO server instance.
        """
        self.sio:SocketIO = sio

    @set_event(message=f"Created", classification="Tag", priority=1, criticity=1)
    def set_tag(
        self, 
        name:str, 
        unit:str, 
        data_type:str, 
        description:str, 
        variable:str,
        display_name:str="",
        display_unit:str="",
        opcua_address:str="",
        node_namespace:str="",
        scan_time:int=None,
        dead_band:float=None,
        process_filter:bool=False,
        gaussian_filter:bool=False,
        gaussian_filter_threshold:float=1.0,
        gaussian_filter_r_value:float=0.0,
        outlier_detection:bool=False,
        out_of_range_detection:bool=False,
        frozen_data_detection:bool=False,
        manufacturer:str="",
        segment:str="",
        id:str=None,
        user:User=None
        )->tuple[Tag, str]:
        """
        Creates and registers a new Tag in the CVT.

        **Parameters:**

        * **name** (str): Unique tag name.
        * **unit** (str): Base unit.
        * **data_type** (str): Data type ('float', 'int', 'bool', 'str').
        * **description** (str): Tag description.
        * **variable** (str): Physical variable type.
        * **display_name** (str, optional): UI display name.
        * **display_unit** (str, optional): UI display unit.
        * **opcua_address** (str, optional): OPC UA server address.
        * **node_namespace** (str, optional): OPC UA node ID.
        * **scan_time** (int, optional): Polling interval.
        * **dead_band** (float, optional): Deadband value.
        * **user** (User, optional): User creating the tag.

        **Returns:**

        * **tuple**: (Tag object, Success message or Error message).
        """
        if isinstance(data_type, str):
        
            if data_type in self.data_types:
                if data_type == "float":
                    value = 0.0
                elif data_type == "int":
                    value = 0
                elif data_type == "str":
                    value = ""
                else:
                    value = False
            
        elif isinstance(data_type, (FloatType, IntegerType, StringType, BooleanType)):

            value = data_type()
            data_type.set(name, value)
            data_type = data_type.__name__
            self.set_data_type(data_type)
        
        has_duplicates, message = self.has_duplicates(name=name, display_name=display_name, opcua_address=opcua_address, node_namespace=node_namespace)
        if has_duplicates:

            return None, message
        
        if not display_unit:

            display_unit = unit

        
        tag = Tag(
            name=name,
            unit=unit,
            data_type=data_type,
            description=description,
            variable=variable,
            display_name=display_name,
            display_unit=display_unit,
            opcua_address=opcua_address,
            node_namespace=node_namespace,
            scan_time=scan_time,
            dead_band=dead_band,
            process_filter=process_filter,
            gaussian_filter=gaussian_filter,
            gaussian_filter_threshold=gaussian_filter_threshold,
            gaussian_filter_r_value=gaussian_filter_r_value,
            outlier_detection=outlier_detection,
            out_of_range_detection=out_of_range_detection,
            frozen_data_detection=frozen_data_detection,
            manufacturer=manufacturer,
            segment=segment,
            id=id
        )
        self._tags[tag.id] = tag

        return tag, message

    @set_event(message=f"Updated", classification="Tag", priority=1, criticity=3)
    def update_tag(
        self, 
        id:str,  
        user:User=None, 
        **kwargs
        )->tuple[Tag|None, str]:
        r"""
        Updates an existing Tag's properties.

        **Parameters:**

        * **id** (str): Tag ID.
        * **user** (User, optional): User performing the update.
        * **kwargs**: Tag properties to update (e.g., name, unit, scan_time).

        **Returns:**

        * **tuple**: (Updated Tag object, Status message).
        """
        check = dict()
        if "name" in kwargs:
            check["name"] = kwargs["name"]
        if "display_name" in kwargs:
            check["display_name"] = kwargs["display_name"]
        if "node_namespace" in kwargs:
            check["node_namespace"] = kwargs["node_namespace"]
        if "opcua_address" in kwargs:
            check["opcua_address"] = kwargs["opcua_address"]
        has_duplicates, message = self.has_duplicates(**check)
        if has_duplicates:

            return None, message
        
        tag = self._tags[id]
        if "name" in kwargs:
            tag.set_name(name=kwargs["name"])
        if "unit" in kwargs:
            tag.set_unit(unit=kwargs["unit"])
        if "data_type" in kwargs:
            tag.set_data_type(data_type=kwargs["data_type"])
        if "description" in kwargs:
            tag.set_description(description=kwargs["description"])
        if "variable" in kwargs:
            tag.set_variable(variable=kwargs["variable"])
        if "display_name" in kwargs:
            tag.set_display_name(name=kwargs["display_name"])
        if "display_unit" in kwargs:
            tag.set_display_unit(unit=kwargs["display_unit"])
        if "opcua_address" in kwargs:
            tag.set_opcua_address(opcua_address=kwargs["opcua_address"])
        if "node_namespace" in kwargs:
            tag.set_node_namespace(node_namespace=kwargs["node_namespace"])
        if "scan_time" in kwargs:
            if isinstance(kwargs["scan_time"], int):
                tag.set_scan_time(scan_time=kwargs["scan_time"])
        if "dead_band" in kwargs:
            tag.set_dead_band(dead_band=kwargs["dead_band"])
        if "segment" in kwargs:
            tag.segment = kwargs["segment"]
        if "manufacturer" in kwargs:
            tag.manufacturer = kwargs["manufacturer"]
        if "gaussian_filter" in kwargs:
            
            if kwargs['gaussian_filter'].lower() in ('1', 'true'):

                tag.gaussian_filter = True

            else:

                tag.gaussian_filter = False

        if "gaussian_filter_r_value" in kwargs:

            tag.gaussian_filter_r_value = kwargs['gaussian_filter_r_value']

        if "gaussian_filter_threshold" in kwargs:

            tag.gaussian_filter_threshold = kwargs['gaussian_filter_threshold']
        
        self._tags[id] = tag

        return tag, f"Tag: {tag.name}"

    @set_event(message=f"Updated", classification="Tag", priority=1, criticity=5)
    def delete_tag(self, id:str, user:User):
        r"""
        Removes a tag from the CVT.

        **Parameters:**

        * **id** (str): Tag ID.
        * **user** (User): User performing the deletion.

        **Returns:**

        * **tuple**: (Deleted Tag object, Status message).
        """
        tag = self._tags.pop(id)
        return tag, f"Tag: {tag.name}"

    def get_tag(self, id:str)->Tag|None:
        r"""
        Retrieves a tag by its ID.

        **Parameters:**

        * **id** (str): Tag ID.

        **Returns:**

        * **Tag**: The Tag object if found, else None.
        """
        for _id, tag in self._tags.items():

            if _id==id:
                
                return tag

        return None
    
    def get_unit_by_tag(self, tag:str)->Tag|None:
        r"""
        Gets the base unit of a tag by name.

        **Parameters:**

        * **tag** (str): Tag name.

        **Returns:**

        * **str**: Unit symbol or None.
        """
        for _id, _tag in self._tags.items():

            if _tag.name==tag:
                
                return _tag.unit

        return None
    
    def get_display_unit_by_tag(self, tag:str)->Tag|None:
        r"""
        Gets the display unit of a tag by name.

        **Parameters:**

        * **tag** (str): Tag name.

        **Returns:**

        * **str**: Display unit symbol or None.
        """
        for _id, _tag in self._tags.items():
            
            if _tag.name==tag:
                
                return _tag.display_unit

        return None

    def get_tags(self)->list:
        r"""
        Returns a list of all tags (serialized).

        **Returns:**

        * **list**: List of tag dictionaries.
        """
        if self._tags:

            return [tag.serialize() for _, tag in self._tags.items()]
        
        return list()
    
    def get_tags_by_names(self, names:list)->list:
        r"""
        Returns a list of serialized tags filtering by name.

        **Parameters:**

        * **names** (list): List of tag names.

        **Returns:**

        * **list**: List of tag dictionaries.
        """
        if self._tags:

            return [tag.serialize() for _, tag in self._tags.items() if tag.name in names]
        
        return list()
    
    def get_field_tags_names(self)->list:
        r"""
        Returns names of tags that are connected to field devices (have OPC UA address and namespace).

        **Returns:**

        * **list**: List of tag names.
        """
        if self._tags:

            return [tag.name for _, tag in self._tags.items() if tag.opcua_address and tag.node_namespace]
        
        return list()
    
    def get_cuasi_field_tags_names(self)->list:
        r"""
        Returns names of tags that have an OPC UA address (potentially field tags).

        **Returns:**

        * **list**: List of tag names.
        """
        if self._tags:

            return [tag.name for _, tag in self._tags.items() if tag.opcua_address]
        
        return list()
    
    def get_tag_by_name(self, name:str)->Tag|None:
        r"""
        Retrieves a tag object by its name.

        **Parameters:**

        * **name** (str): Tag name.

        **Returns:**

        * **Tag**: Tag object or None.
        """
        for _, tag in self._tags.items():
            
            if tag.get_name()==name:
                
                return tag

        return None
    
    def get_tag_by_display_name(self, display_name:str)->Tag|None:
        r"""
        Retrieves a tag object by its display name.

        **Parameters:**

        * **display_name** (str): Tag display name.

        **Returns:**

        * **Tag**: Tag object or None.
        """
        for _, tag in self._tags.items():

            if tag.get_display_name()==display_name:
                
                return tag

        return None

    def get_tag_by_node_namespace(self, node_namespace:str)->Tag|None:
        r"""
        Retrieves a tag object by its OPC UA node namespace.

        **Parameters:**

        * **node_namespace** (str): Node Namespace ID.

        **Returns:**

        * **Tag**: Tag object or None.
        """
        for _, tag in self._tags.items():

            if tag.get_node_namespace()==node_namespace:
                
                return tag

        return None
    
    def get_value(self, id:str)->str|float|int|bool:
        r"""
        Gets the current value of a tag by ID.

        **Parameters:**

        * **id** (str): Tag ID.

        **Returns:**

        * **value**: The current value.
        """
        tag = self._tags[id]        
        _new_object = copy.copy(tag.get_value())
        return _new_object
    
    def get_timestamp(self, id:str)->datetime:
        r"""
        Gets the timestamp of a tag by ID.

        **Parameters:**

        * **id** (str): Tag ID.

        **Returns:**

        * **datetime**: Last update timestamp.
        """
        tag = self._tags[id] 

        return tag.get_timestamp()
    
    def get_value_by_name(self, name:str)->str|float|int|bool:
        r"""
        Gets the value, unit, and timestamp of a tag by name.

        **Parameters:**

        * **name** (str): Tag name.

        **Returns:**

        * **dict**: Dictionary with 'value', 'unit', and 'timestamp'.
        """

        tag = self.get_tag_by_name(name=name)  

        return {
                "value": tag.get_value(),
                "unit": tag.get_unit(),
                "timestamp": tag.get_timestamp()
            }

    def get_values_by_name(self, names:list[str])->str|float|int|bool:
        r"""
        Gets values for multiple tags by name.

        **Parameters:**

        * **names** (list): List of tag names.

        **Returns:**

        * **dict**: Dictionary mapping tag names to their values and metadata.
        """
        data = dict()

        for name in names:

            tag = self.get_tag_by_name(name=name)  
            data[name] = {
                "value": tag.get_value(),
                "unit": tag.get_unit(),
                "timestamp": tag.get_timestamp()
            }
            
        return data
    
    @logging_error_handler
    @filter
    # @iad_frozen_data
    # @iad_out_of_range
    # @iad_outlier
    def set_value(self, id:str, value, timestamp:datetime):
        """
        Sets a new value for a tag.

        This method applies filters (like deadband) and emits a socket event if configured.

        **Parameters:**

        * **id** (str): Tag ID.
        * **value**: New value.
        * **timestamp** (datetime): Timestamp of the value.

        **Returns:**

        * **value**: The value that was set (or filtered).
        """
        from .. import TIMEZONE
        tag = self._tags[id]
        
        # Deadband Logic Wrapper for CVT
        

        if tag.dead_band and isinstance(value, (int, float)):
            try:
                current_value = tag.value.value
                if abs(value - current_value) < tag.dead_band:
                    return value
            except Exception as e:
                logging.error(f"Error in deadband logic: {e}")

        tag.set_value(value=value, timestamp=timestamp)
        if self.sio:
            timestamp = timestamp.astimezone(TIMEZONE)
            self._tags[id].timestamp = timestamp
            self.sio.emit("on.tag", data=self._tags[id].serialize())

        return value

    def set_data_type(self, data_type):
        r"""
        Registers a new data type in the allowed types list.

        **Parameters:**

        * **data_type**: Data type definition.
        """
        self.data_types.append(data_type)
        self.data_types = list(set(self.data_types))
    
    def is_tag_defined(self, name:str)->bool:
        r"""
        Checks if a tag is defined in the CVT.

        **Parameters:**

        * **name** (str): Tag name.

        **Returns:**

        * **bool**: True if defined, False otherwise.
        """

        return name in self._tags
    
    def attach_observer(self, name, observer):
        r"""
        Attaches a new observer to a tag object defined by name.
        
        **Parameters:**

        * **name** (str): Tag name.
        * **observer** (TagObserver): Observer object.
        """
        tag = self.get_tag_by_name(name)
        if tag:
            
            self._tags[tag.id].attach(observer)
        
        else:
            logger = logging.getLogger("pyautomation")
            logger.warning(f"{name} tag Not exists in CVT.attach_observer method")

    def detach_observer(self, name, observer):
        r"""
        Detaches an observer from a tag object defined by name.
        
        **Parameters:**

        * **name** (str): Tag name.
        * **observer** (TagObserver): Observer object.
        """
        tag = self.get_tag_by_name(name)
        self._tags[tag.id].detach(observer)

    def has_duplicates(self, tag:Tag=None, name:str=None, display_name:str=None, node_namespace:str=None, opcua_address:str=None):
        r"""
        Checks for duplicate tag definitions.

        **Parameters:**

        * **tag** (Tag, optional): Existing tag object to compare against.
        * **name** (str, optional): Name to check.
        * **display_name** (str, optional): Display name to check.
        * **node_namespace** (str, optional): Node namespace to check.
        * **opcua_address** (str, optional): OPC UA address to check.

        **Returns:**

        * **tuple**: (Has Duplicates bool, Message str).
        """

        for _, _tag in self._tags.items():

            if name:

                if _tag.get_name()==name:

                    return True, f"Duplicated Tag Name: {name}"
                
            if display_name:
            
                if _tag.get_display_name()==display_name:

                    return True, f"Duplicated Display Name: {display_name}"
                
            if node_namespace:
            
                if _tag.get_node_namespace()==node_namespace:
                    
                    if tag:
                    
                        if _tag.get_opcua_address()==tag.get_opcua_address():

                            return True, f"Duplicated Node Namespace: {node_namespace}"
                        
                    return True, f"Duplicated Node Namespace: {node_namespace}"
            
        return False, f"Valid Tag Name: {name} - Display Name: {display_name}"

    def serialize(self, id:str)->dict:
        r"""
        Serializes a tag by ID.

        **Parameters:**

        * **id** (str): Tag ID.

        **Returns:**

        * **dict**: Serialized tag.
        """
        return self._tags[id].serialize()
    
    def serialize_by_tag_name(self, name:str)->dict|None:
        r"""
        Serializes a tag by Name.

        **Parameters:**

        * **name** (str): Tag name.

        **Returns:**

        * **dict**: Serialized tag.
        """
        tag = self.get_tag_by_name(name)

        if tag:

            return tag.serialize()


class CVTEngine(Singleton):
    """
    Current Value Table (CVT) Engine class for a tag-based, thread-safe repository.

    This class is designed to hold in-memory tag-based values and manage observers for the required tags. It is implemented as a singleton, ensuring that each sub-thread within the PyAutomation application can access and modify tags in a thread-safe manner.

    It acts as a thread-safe wrapper around the `CVT` class, using a query-based mechanism (`request`/`response`) to handle operations.

    **Usage Example**:

    .. code-block:: python

        >>> from automation.tags import CVTEngine
        >>> tag_engine = CVTEngine()
    """


    def __init__(self):

        super(CVTEngine, self).__init__()
        self._cvt = CVT()
        self._request_lock = threading.Lock()
        self._response_lock = threading.Lock()
        self._config = None
        self._response = None
        self._response_lock.acquire()
        self.DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"

    def set_tag(
        self, 
        name:str, 
        unit:str, 
        data_type:str, 
        variable:str,
        description:str, 
        display_unit:str="",
        display_name:str="",
        opcua_address:str="",
        node_namespace:str="",
        scan_time:int=0,
        dead_band:float=0.0,
        process_filter:bool=False,
        gaussian_filter:bool=False,
        gaussian_filter_threshold:float=1.0,
        gaussian_filter_r_value:float=0.0,
        outlier_detection:bool=False,
        out_of_range_detection:bool=False,
        frozen_data_detection:bool=False,
        manufacturer:str="",
        segment:str="",
        id:str="",
        user:User|None=None
        )->tuple[Tag, str]:
        r"""
        Thread-safe method to create a new tag.

        See `CVT.set_tag` for parameters.
        """
        _query = dict()
        _query["action"] = "set_tag"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["unit"] = unit
        _query["parameters"]["data_type"] = data_type
        _query["parameters"]["variable"] = variable
        _query["parameters"]["description"] = description
        _query["parameters"]["display_unit"] = display_unit
        _query["parameters"]["display_name"] = display_name
        _query["parameters"]["opcua_address"] = opcua_address
        _query["parameters"]["node_namespace"] = node_namespace
        _query["parameters"]["scan_time"] = scan_time
        _query["parameters"]["dead_band"] = dead_band
        _query["parameters"]["process_filter"] = process_filter
        _query["parameters"]["gaussian_filter"] = gaussian_filter
        _query["parameters"]["gaussian_filter_threshold"] = gaussian_filter_threshold
        _query["parameters"]["gaussian_filter_r_value"] = gaussian_filter_r_value
        _query["parameters"]["outlier_detection"] = outlier_detection
        _query["parameters"]["out_of_range_detection"] = out_of_range_detection
        _query["parameters"]["frozen_data_detection"] = frozen_data_detection
        _query["parameters"]["manufacturer"] = manufacturer
        _query["parameters"]["segment"] = segment
        _query["parameters"]["id"] = id
        _query["parameters"]["user"] = user
        return self.__query(_query)
    
    def update_tag(
            self, 
            id:str,  
            user:User=None, 
            **kwargs
        ):
        r"""
        Thread-safe method to update a tag.

        See `CVT.update_tag` for parameters.
        """
        _query = dict()
        _query["action"] = "update_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["user"] = user
        for key, value in kwargs.items():

            _query["parameters"][key] = value
        return self.__query(_query)
    
    def delete_tag(self, id:str, user:User|None=None):
        r"""
        Thread-safe method to delete a tag.

        See `CVT.delete_tag` for parameters.
        """
        _query = dict()
        _query["action"] = "delete_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["user"] = user
        return self.__query(_query)
    
    def get_tag(
        self,
        id:str=None
        )->Tag:
        r"""
        Thread-safe method to get a tag object.
        """
        _query = dict()
        _query["action"] = "get_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)

    def get_tags(self):
        r"""
        Thread-safe method to get all tags.
        """
        _query = dict()
        _query["action"] = "get_tags"
        return self.__query(_query)

    def get_tags_by_names(self, names:list[str]):
        r"""
        Thread-safe method to get tags by names.
        """
        _query = dict()
        _query["action"] = "get_tags_by_names"
        _query["parameters"] = dict()
        _query["parameters"]["names"] = names
        return self.__query(_query)
    
    def get_tag_by_name(self, name:str)->Tag|None:
        r"""
        Thread-safe method to get a tag by name.
        """
        _query = dict()
        _query["action"] = "get_tag_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        return self.__query(_query)
    
    def get_tag_by_display_name(self, display_name:str)->Tag|None:
        r"""
        Thread-safe method to get a tag by display name.
        """
        _query = dict()
        _query["action"] = "get_tag_by_display_name"
        _query["parameters"] = dict()
        _query["parameters"]["display_name"] = display_name
        return self.__query(_query)

    def get_tag_by_node_namespace(self, node_namespace:str)->Tag|None:
        r"""
        Thread-safe method to get a tag by node namespace.
        """
        _query = dict()
        _query["action"] = "get_tag_by_node_namespace"
        _query["parameters"] = dict()
        _query["parameters"]["node_namespace"] = node_namespace
        return self.__query(_query)
    
    def get_value(self, id:str)->str|float|int|bool:
        r"""
        Thread-safe method to get a tag value by ID.
        """
        _query = dict()
        _query["action"] = "get_value"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)
    
    def get_value_by_name(self, tag_name:str)->dict:
        r"""
        Thread-safe method to get a tag value by name.
        """
        _query = dict()
        _query["action"] = "get_value_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = tag_name
        return self.__query(_query)
    
    def get_values_by_name(self, tag_names:list[str])->str|float|int|bool:
        r"""
        Thread-safe method to get multiple tag values.
        """
        _query = dict()
        _query["action"] = "get_values_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["names"] = tag_names
        return self.__query(_query)
    
    def get_scan_time(self, id:str)->str|float|int|bool:
        r"""
        Thread-safe method to get scan time.
        """
        _query = dict()
        _query["action"] = "get_scan_time"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)
    
    def get_dead_band(self, id:str)->str|float|int|bool:
        r"""
        Thread-safe method to get deadband.
        """
        _query = dict()
        _query["action"] = "get_dead_band"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)
    
    def get_display_unit_by_tag(self, tag:str)->str:
        r"""
        Thread-safe method to get display unit.
        """
        _query = dict()
        _query["action"] = "get_display_unit_by_tag"
        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        return self.__query(_query)
    
    @logging_error_handler
    def set_value(self, id:str, value, timestamp:datetime):
        r"""
        Thread-safe method to set a tag value.
        """
        if not timestamp:
            timestamp = datetime.now()
        _query = dict()
        _query["action"] = "set_value"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["value"] = value
        _query["parameters"]["timestamp"] = timestamp
        return self.__query(_query)
    
    def set_data_type(self, data_type):
        r"""
        Thread-safe method to set data type.
        """
        _query = dict()
        _query["action"] = "set_data_type"
        _query["parameters"] = dict()
        _query["parameters"]["data_type"] = data_type
        return self.__query(_query)
    
    def is_tag_defined(self, name:str)->bool:
        r"""
        Thread-safe method to check if tag is defined.
        """
        _query = dict()
        _query["action"] = "is_tag_defined"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        return self.__query(_query)

    def attach(self, name:str, observer):
        """
        Attaches an observer to a Tag in a thread-safe way.
        """
        _query = dict()
        _query["action"] = "attach_observer"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["observer"] = observer
        return self.__query(_query)

    def detach(self, name:str, observer):
        """
        Detaches an observer from a Tag in a thread-safe way.
        """
        
        _query = dict()
        _query["action"] = "detach_observer"

        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["observer"] = observer

        self.request(_query)
        result = self.response()

        if result["result"]:
            return result["response"]

    def serialize(self, id:str)->dict:
        r"""
        Thread-safe serialization by ID.
        """
        _query = dict()
        _query["action"] = "serialize"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)
    
    def serialize_by_tag_name(self, name:str)->dict|None:
        r"""
        Thread-safe serialization by name.
        """
        _query = dict()
        _query["action"] = "serialize_by_tag_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        return self.__query(_query)

    def __query(self, query:dict)->dict:

        self.request(query)
        result = self.response()
        if result["result"]:
            return result["response"]

    def request(self, query:dict):
        r"""
        Executes a request to the CVT in a thread-safe mechanism using locks.

        **Parameters:**

        * **query** (dict): Dictionary defining the action and parameters.
        """
        self._request_lock.acquire()
        action = query["action"]
        error_msg = f"Error in CVTEngine with action: {action}"

        try:

            if hasattr(self._cvt, action):

                method = getattr(self._cvt, action)
                
                if 'parameters' in query:
                    
                    resp = method(**query["parameters"])

                else:

                    resp = method()

            self.__true_response(resp)

        except Exception as e:
            
            self.__log_error(e, error_msg)

        self._response_lock.release()

    def __log_error(self, e:Exception, msg:str):
        r"""
        Logs error and sets error response.
        """
        logging.error(f"{e} Message: {msg}")
        self._response = {
            "result": False,
            "response": None
        }

    def __true_response(self, resp):
        r"""
        Sets success response.
        """
        self._response = {
            "result": True,
            "response": resp
        }

    def response(self)->dict:
        r"""
        Retrieves the response from the last request, handling thread synchronization.
        """
        self._response_lock.acquire()

        result = self._response

        self._request_lock.release()

        return result

    def __getstate__(self):

        self._response_lock.release()
        state = self.__dict__.copy()
        del state['_request_lock']
        del state['_response_lock']
        return state

    def __setstate__(self, state):
        
        self.__dict__.update(state)
        self._request_lock = threading.Lock()
        self._response_lock = threading.Lock()
        self._response_lock.acquire()



# -*- coding: utf-8 -*-
"""automation/logger/events.py

This module implements the Events Logger, responsible for persisting system events
such as user actions, system notifications, and critical alerts to the database.
"""
from datetime import datetime
from ..dbmodels.events import Events
from ..modules.users.users import User
from .core import BaseEngine, BaseLogger


class EventsLogger(BaseLogger):
    r"""
    Logger class specialized for Event Management.

    It handles the creation and retrieval of system events, including filtering capabilities.
    """

    def __init__(self):

        super(EventsLogger, self).__init__()

    def create(
        self, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        priority:int=None,
        criticity:int=None,
        timestamp:datetime=None
        ):
        r"""
        Creates a new event record in the database.

        **Parameters:**

        * **message** (str): The event message.
        * **user** (User): The user associated with the event.
        * **description** (str, optional): Detailed description.
        * **classification** (str, optional): Event category.
        * **priority** (int, optional): Priority level.
        * **criticity** (int, optional): Criticality level.
        * **timestamp** (datetime, optional): Time of the event.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return None
            
        return Events.create(
            message=message, 
            user=user, 
            description=description, 
            classification=classification,
            priority=priority,
            criticity=criticity,
            timestamp=timestamp
        )

    def get_lasts(self, lasts:int=1):
        r"""
        Retrieves the most recent events.

        **Parameters:**

        * **lasts** (int): Number of events to retrieve.

        **Returns:**

        * **list**: List of Events.
        """
        if not self.is_history_logged:

            return list()
        
        if not self.check_connectivity():
            
            return list()
        
        return Events.read_lasts(lasts=lasts)
    
    def filter_by(
        self,
        usernames:list[str]=None,
        priorities:list[int]=None,
        criticities:list[int]=None,
        message:str="",
        description:str="",
        classification:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str="UTC",
        page:int=1,
        limit:int=20
        ):
        r"""
        Filters events based on multiple criteria.

        **Parameters:**

        * **usernames** (list[str]): Filter by usernames.
        * **priorities** (list[int]): Filter by priority.
        * **criticities** (list[int]): Filter by criticality.
        * **message** (str): Partial match on message.
        * **description** (str): Partial match on description.
        * **classification** (str): Partial match on classification.
        * **greater_than_timestamp** (datetime): Start time.
        * **less_than_timestamp** (datetime): End time.
        * **timezone** (str): Timezone.
        * **page** (int): Page number.
        * **limit** (int): Records per page.

        **Returns:**

        * **dict**: Filtered results with pagination metadata.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
        
        return Events.filter_by(
            usernames=usernames,
            priorities=priorities,
            criticities=criticities,
            message=message,
            classification=classification,
            description=description,
            greater_than_timestamp=greater_than_timestamp,
            less_than_timestamp=less_than_timestamp,
            timezone=timezone,
            page=page,
            limit=limit
            )

    def get_summary(self)->tuple[list, str]:
        r"""
        Retrieves a summary of all events.

        **Returns:**

        * **list**: List of serialized event dictionaries.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
            
        return Events.serialize()
    
class EventsLoggerEngine(BaseEngine):
    r"""
    Thread-safe Engine for the EventsLogger.
    """

    def __init__(self):

        super(EventsLoggerEngine, self).__init__()
        self.logger = EventsLogger()

    def create(
        self,
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        priority:int=None,
        criticity:int=None,
        timestamp:datetime=None
        ):
        r"""
        Thread-safe event creation.
        """
        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["message"] = message
        _query["parameters"]["user"] = user
        _query["parameters"]["description"] = description
        _query["parameters"]["classification"] = classification
        _query["parameters"]["priority"] = priority
        _query["parameters"]["criticity"] = criticity
        _query["parameters"]["timestamp"] = timestamp
        
        return self.query(_query)
    
    def get_lasts(
        self,
        lasts:int=1
        ):
        r"""
        Thread-safe retrieval of last events.
        """
        _query = dict()
        _query["action"] = "get_lasts"
        _query["parameters"] = dict()
        _query["parameters"]["lasts"] = lasts
        
        return self.query(_query)
    
    def filter_by(
        self,
        usernames:list[str]=None,
        priorities:list[int]=None,
        criticities:list[int]=None,
        message:str="",
        classification:str="",
        description:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):
        r"""
        Thread-safe event filtering.
        """
        _query = dict()
        _query["action"] = "filter_by"
        _query["parameters"] = dict()
        _query["parameters"]["usernames"] = usernames
        _query["parameters"]["priorities"] = priorities
        _query["parameters"]["criticities"] = criticities
        _query["parameters"]["message"] = message
        _query["parameters"]["classification"] = classification
        _query["parameters"]["description"] = description
        _query["parameters"]["greater_than_timestamp"] = greater_than_timestamp
        _query["parameters"]["less_than_timestamp"] = less_than_timestamp
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["page"] = page
        _query["parameters"]["limit"] = limit
        
        return self.query(_query)

    def get_summary(self):
        r"""
        Thread-safe retrieval of event summary.
        """
        _query = dict()
        _query["action"] = "get_summary"
        _query["parameters"] = dict()
        
        return self.query(_query)



from flask_restx import Namespace, Resource, fields
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api

ns = Namespace('Settings', description='Application Configuration Settings')
app = PyAutomation()

settings_model = api.model("settings_model", {
    'logger_period': fields.Float(required=False, min=1.0, description='Logger worker period in seconds (>= 1.0)'),
    'log_max_bytes': fields.Integer(required=False, min=1024, description='Max bytes for log file rotation (>= 1024)'),
    'log_backup_count': fields.Integer(required=False, min=1, description='Number of backup log files to keep (>= 1)'),
    'log_level': fields.Integer(required=False, min=0, max=50, description='Logging level (0=NOTSET, 10=DEBUG, 20=INFO, 30=WARNING, 40=ERROR, 50=CRITICAL)')
})

@ns.route('/update')
class SettingsUpdateResource(Resource):
    
    @api.doc(security='apikey', description="Updates various application settings.")
    @api.response(200, "Settings updated successfully")
    @api.response(400, "Invalid parameter values")
    @Api.token_required(auth=True)
    @ns.expect(settings_model)
    def put(self):
        """
        Update settings.

        Updates application configuration including logger period, log rotation settings, and logging level.
        """
        data = api.payload
        
        # 1. Update Logger Worker Period
        if 'logger_period' in data:
            logger_period = data['logger_period']
            if logger_period < 1.0:
                return "Logger period must be >= 1.0", 400
            app.update_logger_period(logger_period)

        # 2. Update Log Rotation Config
        if 'log_max_bytes' in data and 'log_backup_count' in data:
            max_bytes = data['log_max_bytes']
            backup_count = data['log_backup_count']
            
            if max_bytes < 1024:
                 return "log_max_bytes must be >= 1024", 400
            if backup_count < 1:
                 return "log_backup_count must be >= 1", 400
                 
            app.update_log_config(max_bytes, backup_count)
        
        elif 'log_max_bytes' in data or 'log_backup_count' in data:
             return "Both log_max_bytes and log_backup_count must be provided together", 400

        # 3. Update Log Level
        if 'log_level' in data:
            log_level = data['log_level']
            # Basic validation for standard levels
            if log_level not in [0, 10, 20, 30, 40, 50]:
                return "Invalid log_level. Use standard Python logging levels (10, 20, 30, 40, 50)", 400
            
            app.update_log_level(log_level)

        return "Settings updated", 200

@ns.route('/logger_period')
class LoggerPeriodResource(Resource):
    
    @api.doc(security='apikey', description="Updates only the logger worker period.")
    @api.response(200, "Logger period updated successfully")
    @api.response(400, "Invalid logger period")
    @Api.token_required(auth=True)
    @ns.expect(settings_model)
    def put(self):
        """
        Update logger period.

        Updates the execution period of the logger worker.
        """
        logger_period = api.payload.get('logger_period')
        
        if logger_period < 1.0:
            return "Logger period must be >= 1.0", 400
            
        app.update_logger_period(logger_period)
        
        return "Logger period updated", 200



import secrets
from peewee import CharField, IntegerField, ForeignKeyField
from ..dbmodels.core import BaseModel
from ..modules.users.users import Users as CVTUsers
from ..modules.users.roles import Roles as CVTRoles
from ..modules.users.roles import Role
from ..modules.users.users import User
from werkzeug.security import generate_password_hash, check_password_hash

users = CVTUsers()
roles = CVTRoles()

class Roles(BaseModel):
    r"""
    Database model for User Roles.
    """
    
    identifier = CharField(unique=True, max_length=16)
    name = CharField(unique=True, max_length=32)
    level = IntegerField()
    __defaults__ = [
        {"name": "sudo", "level": 0, "identifier": secrets.token_hex(4)},
        {"name": "admin", "level": 1, "identifier": secrets.token_hex(4)},
        {"name": "supervisor", "level": 2, "identifier": secrets.token_hex(4)},
        {"name": "operator", "level": 10, "identifier": secrets.token_hex(4)},
        {"name": "auditor", "level": 100, "identifier": secrets.token_hex(4)},
        {"name": "guest", "level": 256, "identifier": secrets.token_hex(4)}
    ]

    @classmethod
    def create(cls, name:str, level:int, identifier:str)->tuple:
        r"""
        Creates a new Role record.

        **Parameters:**

        * **name** (str): Role name.
        * **level** (int): Permission level (0 is highest).
        * **identifier** (str): Unique ID.

        **Returns:**

        * **tuple**: (Query object, status message)
        """
        
        name = name.upper()

        if cls.name_exist(name):

            return None, f"role {name} is already used"

        if cls.identifier_exist(identifier):
                
            return None, f"identifier {identifier} is already used"
        
        query = cls(name=name, level=level, identifier=identifier)
        query.save()
        
        return query, f"Role creation successful"

    @classmethod
    def read_by_name(cls, name:str):
        r"""
        Retrieves a role by name.
        """
      
        return cls.get_or_none(name=name.upper())
    
    @classmethod
    def read_by_identifier(cls, identifier:str):
        r"""
        Retrieves a role by identifier.
        """
      
        return cls.get_or_none(identifier=identifier)

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Checks if a role name exists.
        """
        
        return True if cls.get_or_none(name=name.upper()) else False
    
    @classmethod
    def identifier_exist(cls, identifier:str)->bool:
        r"""
        Checks if a role identifier exists.
        """

        return True if cls.get_or_none(identifier=identifier) else False
    
    @classmethod
    def read_names(cls)->list:
        r"""
        Returns a list of all role names.
        """

        return [role.name for role in cls.select()]
    
    @classmethod
    def fill_cvt_roles(cls):
        r"""
        Loads roles from the database into the in-memory Role Manager (CVT).
        """
        for role in cls.select():

            _role = Role(
                name=role.name,
                level=role.level,
                identifier=role.identifier
            )

            roles.add(role=_role)

    def serialize(self)->dict:
        r"""
        Serializes the role record.
        """

        return {
            "id": self.id,
            "identifier": self.identifier,
            "name": self.name,
            "level": self.level
        }


class Users(BaseModel):
    r"""
    Database model for Users.
    """
    
    identifier = CharField(unique=True, max_length=16)
    username = CharField(unique=True, max_length=64)
    role = ForeignKeyField(Roles, backref='users', on_delete='CASCADE')
    email = CharField(unique=True, max_length=128)
    password = CharField()
    token = CharField(null=True)
    name = CharField(max_length=64, null=True)
    lastname = CharField(max_length=64, null=True)

    @classmethod
    def create(cls, user:User)-> dict:
        r"""
        Creates a new User record.

        **Parameters:**

        * **user** (User): An instance of the User class from the users module.

        **Returns:**

        * **tuple**: (Query object, status message)
        """

        if cls.username_exist(user.username):

            return None, f"username {user.username} is already used"

        if cls.email_exist(user.email):

            return None, f"email {user.email} is already used"
        
        if cls.identifier_exist(user.identifier):

            return None, f"identifier {user.identifier} is already used"
        
        query = cls(
            username=user.username,
            role=Roles.read_by_name(name=user.role.name),
            email=user.email,
            password=user.password,
            identifier=user.identifier,
            name=user.name,
            lastname=user.lastname,
            token=user.token
            )
        query.save()

        return query, f"User creation successful"
    
    @classmethod
    def login(cls, password:str, username:str="", email:str=""):
        r"""
        Authenticates a user and updates their session token.

        **Parameters:**

        * **password** (str): User password.
        * **username** (str, optional): Username.
        * **email** (str, optional): Email.

        **Returns:**

        * **tuple**: (User record, status message)
        """
        if username:

            user = cls.get_or_none(username=username)

        if email:

            user = cls.get_or_none(email=email)
        
        if user:

            if user.decode_password(password):
                
                if not user.token:
                    
                    user.token = cls.encode(secrets.token_hex(4))
                    user.save()
                
                users.login(password=password, token=user.token, username=username, email=email)

                return user, f"Login successfull"

            return None, f"Invalid credentials" 

        return None, f"Invalid Username or Email"      
    
    @classmethod
    def logout(cls, token:str):
        r"""
        Logs out a user by invalidating their token.

        **Parameters:**

        * **token** (str): Session token.
        """
        user = cls.get_or_none(token=token)

        if user:

            user.token = None
            user.save()

            return user, "Logout successfull"
        
        return None, "Invalid Token"

    @classmethod
    def read_by_username(cls, username:str):
        r"""
        Retrieves a user by username.
        """
   
        return cls.get_or_none(username=username)

    @classmethod
    def read_by_name(cls, name:str):
        r"""
        Retrieves a user by name (first name).
        """
   
        return cls.get_or_none(name=name)

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Checks if a user name exists.
        """

        return True if cls.get_or_none(name=name) else False
    
    @classmethod
    def username_exist(cls, username:str)->bool:
        r"""
        Checks if a username exists.
        """

        return True if cls.get_or_none(username=username) else False
    
    @classmethod
    def email_exist(cls, email:str)->bool:
        r"""
        Checks if an email exists.
        """

        return True if cls.get_or_none(email=email) else False
    
    @classmethod
    def identifier_exist(cls, identifier:str)->bool:
        r"""
        Checks if an identifier exists.
        """

        return True if cls.get_or_none(identifier=identifier) else False
    
    @classmethod
    def encode(cls, value:str)->str:
        r"""
        Hashes a value (e.g., password).
        """

        return generate_password_hash(value)
    

    def decode_password(self, password:str)->str:
        r"""
        Verifies a password against the stored hash.
        """

        return check_password_hash(self.password, password)
    
    def decode_token(self, token:str)->str:
        r"""
        Verifies a token against the stored hash.
        """

        return check_password_hash(self.token, token)
    
    @classmethod
    def fill_cvt_users(cls):
        r"""
        Loads users from the database into the in-memory User Manager (CVT).
        """
        for user in cls.select():

            users.signup(
                username=user.username,
                role_name=user.role.name,
                email=user.email,
                password=user.password,
                name=user.name,
                lastname=user.lastname,
                identifier=user.identifier,
                encode_password=False
            )

    def serialize(self)-> dict:
        r"""
        Serializes the user record.
        """

        return {
            "id": self.id,
            "identifier": self.identifier,
            "username": self.username,
            "email": self.email,
            "role": self.role.serialize(),
            "name": self.name,
            "lastname": self.lastname
        }


import dash
import dash_bootstrap_components as dbc
from ...variables import VARIABLES

class TagsComponents:

    @classmethod
    def create_tag_form(cls):
        r"""
        Documentation here
        """
        return dash.html.Div(
            [
                dash.dcc.Location(id='tags_page', refresh=False),
                dbc.Accordion(
                    [
                        dbc.AccordionItem(
                            [
                                dbc.Row(
                                    [
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="Tag Name", id="tag_name_input")
                                                    ], 
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Variable"),
                                                        dbc.Select(
                                                            options=[
                                                                {"label": variable, "value": variable} for variable in VARIABLES.keys()
                                                            ],
                                                            id="variable_input"
                                                        ),
                                                    ],
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Datatype"),
                                                        dbc.Select(
                                                            options=[
                                                                {'label': 'Float', 'value': 'float'},
                                                                {'label': 'Integer', 'value': 'integer'},
                                                                {'label': 'Boolean', 'value': 'boolean'},
                                                                {'label': 'String', 'value': 'string'}
                                                            ],
                                                            id="datatype_input"
                                                        ),
                                                    ],
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-3"
                                        ),
                                        
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Unit"), dbc.Select(options=[], id="unit_input", disabled=True )
                                                    ],
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Display Unit"), dbc.Select(options=[], id="display_unit_input", disabled=True )
                                                    ],
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="Manufacturer (Optional)", id="manufacturer_input", disabled=True)
                                                    ], 
                                                    size="md"
                                                )
                                            
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-2"
                                        ),
                                        
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText(dbc.RadioButton(id="description_radio_button"), class_name="radiobutton-box"), 
                                                        dbc.Input(placeholder="Description (Optional)", id="description_input", disabled=True)
                                                    ], 
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText(dbc.RadioButton(id="display_name_radio_button"), className="radiobutton-box"), 
                                                        dbc.Input(placeholder="Display Name (Optional)", id="display_name_input", disabled=True)
                                                    ], 
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText(dbc.RadioButton(id="segment_radio_button"), className="radiobutton-box"), 
                                                        dbc.Input(placeholder="Segment (Optional)", id="segment_input", disabled=True)
                                                    ], 
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-3"
                                        ),
                                        
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("OPCUA"), 
                                                        dbc.Select(options=[], id="opcua_address_input")
                                                    ],
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.InputGroupText("Node"), 
                                                        dbc.Select(options=[], id="node_namespace_input", disabled=True)
                                                    ],
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-2"
                                        ),
                                        
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="Scan Time", type="number", step=500, min=500, max=600000, id="scan_time_input", disabled=True), 
                                                        dbc.InputGroupText('ms')
                                                    ], 
                                                    size="md",
                                                    className="mb-3"
                                                ),
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="Dead-Band", type="number", step=0.1, id="dead_band_input", disabled=True), 
                                                        dbc.InputGroupText('', id="dead_band_unit")
                                                    ], 
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-2"
                                        ),

                                        dbc.Col(
                                            dbc.Button(
                                                "Create",
                                                color="primary",
                                                outline=True,
                                                disabled=True,
                                                id="create_tag_button",
                                                className="w-100"
                                            ),
                                            width="auto",
                                            className="d-flex justify-content-center align-items-center"
                                        ),
                                    ],
                                    className="form g-3"
                                ),
                            ],
                            title="Create Tag",
                            className="my-3"
                        )
                    ],
                    start_collapsed=True,
                )
            ]
        )

    @classmethod
    def tags_table(cls)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        return dbc.Container(
            dbc.Row(
                dbc.Col(
                    dash.dash_table.DataTable(
                        data=[],
                        columns=[
                            {'name': 'id', 'id': 'id', 'editable': False}, 
                            {'name': 'name', 'id': 'name'}, 
                            {'name': 'variable', 'id': 'variable', 'presentation': 'dropdown', 'clearable': False},
                            {'name': 'unit', 'id': 'unit', 'presentation': 'dropdown'},
                            {'name': 'display_unit', 'id': 'display_unit', 'presentation': 'dropdown'},  
                            {'name': 'data_type', 'id': 'data_type', 'presentation': 'dropdown', 'clearable': False}, 
                            {'name': 'description', 'id': 'description'}, 
                            {'name': 'display_name', 'id': 'display_name'}, 
                            {'name': 'segment', 'id': 'segment', 'presentation': 'dropdown'},
                            {'name': 'opcua_address', 'id': 'opcua_address', 'presentation': 'dropdown'}, 
                            {'name': 'node_namespace', 'id': 'node_namespace', 'presentation': 'dropdown'},
                            {'name': 'scan_time', 'id': 'scan_time', 'type': 'numeric'}, 
                            {'name': 'dead_band', 'id': 'dead_band', 'type': 'numeric'}
                        ],
                        id="tags_datatable",
                        filter_action="native",
                        sort_action="native",
                        sort_mode="multi",
                        row_deletable=True,
                        selected_columns=[],
                        dropdown = {
                            'data_type': {
                                'options': [
                                    {'label': 'Float', 'value': 'float'},
                                    {'label': 'Integer', 'value': 'integer'},
                                    {'label': 'Boolean', 'value': 'boolean'},
                                    {'label': 'String', 'value': 'string'}
                                ],
                                'clearable': False
                            },
                            'opcua_address': {
                                'options': []
                            },
                            'variable': {
                                'options': [{"label": key, "value": key} for key, _ in VARIABLES.items()]
                            }
                        },
                        page_action="native",
                        page_current= 0,
                        page_size= 10,
                        persistence=True,
                        editable=True,
                        persisted_props=['data'],
                        export_format='xlsx',
                        export_headers='display',
                        style_table={'overflowX': 'auto'},
                    ),
                    width=12,
                )
            ),
            fluid=True,
            className="mx-0 px-0"
        )


from ..utils.units import *

class VolumetricFlow(EngUnit):
    """Creates a flow object that can store a flow value and 
    convert between units of flow.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [VolumetricFlow Object]\n

    ```python
    >>> from automation.variables.volumetric_flow import VolumetricFlow
    >>> volumetric_flow = VolumetricFlow(value=1.0, unit="bbl/sec")
    >>> volumetric_flow.value
    1.0
    >>> volumetric_flow.unit
    'bbl/sec'
    >>> volumetric_flow.convert(to_unit="m3/day")
    13736.5022817792
    >>> VolumetricFlow.convert_value(value=1.0, from_unit="gal/sec", to_unit="bbl/sec")
    0.023809523809523808
    >>> VolumetricFlow.convert_values(values=[1.0, 10.0], from_unit="gal/sec", to_unit="bbl/sec")
    [0.023809523809523808, 0.23809523809523808]
    >>> volumetric_flow.change_unit(unit="gal/sec")
    42.0
    >>> volumetric_flow.unit
    'gal/sec'
    >>> volumetric_flow.get_value()
    [42.0, 'gal/sec']
    >>> print(volumetric_flow)
    42.0 gal/sec
    >>> volumetric_flow2 = VolumetricFlow(value=120.0, unit="gal/min")
    >>> volumetric_flow_result = volumetric_flow + volumetric_flow2
    >>> print(volumetric_flow_result)
    44.0 gal/sec
    >>> volumetric_flow_result = volumetric_flow * volumetric_flow2
    >>> print(volumetric_flow_result)
    84.0 gal/sec
    >>> volumetric_flow_result = volumetric_flow / volumetric_flow2
    >>> print(volumetric_flow_result)
    21.0 gal/sec
    >>> volumetric_flow_result = volumetric_flow ** volumetric_flow2
    >>> print(volumetric_flow_result)
    1764.0 gal/sec

    ```
    """

    class Units(UnitSerializer):
        bbl_day = 'bbl/day'
        bbl_hr = 'bbl/hr'
        bbl_min = 'bbl/min'
        bbl_sec = 'bbl/sec'
        gal_day = 'gal/day'
        gal_hr = 'gal/hr'
        gal_min = 'gal/min'
        gal_sec = 'gal/sec'
        cubic_meter_day = 'm3/day'
        cubic_meter_hr = 'm3/hr'
        cubic_meter_min = 'm3/min'
        cubic_meter_sec = 'm3/sec'
        liter_day = 'lt/day'
        liter_hr = 'lt/hr'
        liter_min = 'lt/min'
        liter_sec = 'lt/sec'
        cubic_centimeter_day = 'cc/day'
        cubic_centimeter_hr = 'cc/hr'
        cubic_centimeter_min = 'cc/min'
        cubic_centimeter_sec = 'cc/sec'

    conversions = {
        'bbl/day' : 1.0 * 60.0 * 60.0 * 24.0,
        'bbl/hr' : 1.0 * 60 * 60,
        'bbl/min' : 1.0 * 60.0,
        'bbl/sec' : 1.0,
        'gal/day' : 42.0 * 60.0 * 60.0 * 24.0,
        'gal/hr' : 42.0 * 60.0 * 60.0,
        'gal/min' : 42.0 * 60.0,
        'gal/sec' : 42.0,
        'm3/day' : 0.158987294928 * 60 * 60 * 24,
        'm3/hr' : 0.158987294928 * 60 * 60,
        'm3/min' : 0.158987294928 * 60,
        'm3/sec' : 0.158987294928,
        'lt/day' : 158.987294928 * 60 * 60 * 24,
        'lt/hr' : 158.987294928 * 60 * 60,
        'lt/min' : 158.987294928 * 60,
        'lt/sec' : 158.987294928,
        'ml/day' : 158987.294928 * 60 * 60 * 24,
        'ml/hr' : 158987.294928 * 60 * 60,
        'ml/min' : 158987.294928 * 60,
        'ml/sec' : 158987.294928
    }

    def __init__(self, value, unit):

        if unit not in VolumetricFlow.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {VolumetricFlow.Units.list()}")
        
        super(VolumetricFlow, self).__init__(value=value, unit=unit)


from ..utils.units import EngUnit, UnitError, UnitSerializer

class Current(EngUnit):
    r"""
    Creates a current object pruebaaaaa that can store a current value and 
    convert between units of current.

    :param value: The current value (engineering value)
    :type value: int | float
    :param unit: The unit of the current (e.g., "A", "mA", "kA")
    :type unit: str
    :return: A new Current object
    :rtype: Current

    Example usage:

    .. code-block:: python

        >>> from automation.variables.current import Current
        >>> current = Current(value=1.0, unit="A")
        >>> current.value
        1.0
        >>> current.unit
        'A'
        >>> current.convert(to_unit="kA")
        0.001
        >>> Current.convert_value(value=1.0, from_unit="A", to_unit="mA")
        1000.0
        >>> Current.convert_values(values=[1.0, 10.0], from_unit="A", to_unit="mA")
        [1000.0, 10000.0]
        >>> current.change_unit(unit="kA")
        0.001
        >>> current.unit
        'kA'
        >>> current.get_value()
        [0.001, 'kA']
        >>> print(current)
        0.001 kA
        >>> current2 = Current(value=3.0, unit="mA")
        >>> current_result = current + current2
        >>> print(current_result)
        0.001003 kA
        >>> current_result = current * current2
        >>> print(current_result)
        3e-09 kA
        >>> current_result = current / current2
        >>> print(current_result)
        333.3333333333333 kA
        >>> current_result = current ** current2
        >>> print(current_result)
        0.9999792769488884 kA
    """
    
    class Units(UnitSerializer):
        A = 'A'
        mA = 'mA'
        kA = 'kA'
    
    conversions = {
        'A' : 1,
        'mA' : 1000,
        'kA' : 0.001
    }

    def __init__(self, value, unit):

        if unit not in Current.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Current.Units.list()}")
        
        super(Current, self).__init__(value=value, unit=unit)


from ..utils.units import *

class Density(EngUnit):
    """Creates a density object that can store a density value and 
    convert between units of density.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Density Object]\n

    ```python
    >>> from automation.variables.density import Density
    >>> density = Density(value=1.0, unit="kg/m3")
    >>> density.value
    1.0
    >>> density.unit
    'kg/m3'
    >>> density.convert(to_unit="lb/ml")
    2.2046226218000004e-06
    >>> Density.convert_value(value=1.0, from_unit="kg/m3", to_unit="g/ml")
    0.0010000000000000002
    >>> Density.convert_values(values=[1.0, 10.0], from_unit="kg/m3", to_unit="g/ml")
    [0.0010000000000000002, 0.010000000000000002]
    >>> density.change_unit(unit="lb/ml")
    2.2046226218000004e-06
    >>> density.unit
    'lb/ml'
    >>> density.get_value()
    [2.2046226218000004e-06, 'lb/ml']
    >>> print(density)
    2.2046226218000004e-06 lb/ml
    >>> density2 = Density(value=3.0, unit="g/ml")
    >>> density_result = density + density2
    >>> print(density_result)
    0.0066160724880218 lb/ml
    >>> density_result = density * density2
    >>> print(density_result)
    1.458108271365692e-08 lb/ml
    >>> density_result = density / density2
    >>> print(density_result)
    0.00033333333333333343 lb/ml
    >>> density_result = density ** density2
    >>> print(density_result)
    0.9174608905185654 lb/ml

    ```

    """
    
    class Units(UnitSerializer):
        kg_bbl = 'kg/bbl'
        kg_gal = 'kg/gal'
        kg_m3 = 'kg/m3'
        kg_lt = 'kg/lt'
        kg_ml = 'kg/ml'
        g_bbl = 'g/bbl'
        g_gal = 'g/gal'
        g_m3 = 'g/m3'
        g_lt = 'g/lt'
        g_ml = 'g/ml'
        mg_bbl = 'mg/bbl'
        mg_gal = 'mg/gal'
        mg_m3 = 'kmg/m3'
        mg_lt = 'mg/lt'
        mg_ml = 'mg/ml'
        lb_bbl = 'lb/bbl'
        lb_gal = 'lb/gal'
        lb_m3 = 'lb/m3'
        lb_lt = 'lb/lt'
        lb_ml = 'lb/ml'
        metricTon_bbl = 'metricTon/bbl'
        metricTon_gal = 'metricTon/gal'
        metricTon_m3 = 'metricTon/m3'
        metricTon_lt = 'metricTon/lt'
        metricTon_ml = 'metricTon/ml'
    
    conversions = {
        'kg/bbl' : 1.0,
        'g/bbl' : 1.0 * 1000,
        'mg/bbl' : 1.0 * 1000 * 1000,
        'metricTon/bbl' : 1.0 / 1000.0,
        'lb/bbl' : 2.2046226218,

        'kg/gal' : 1.0 / 42.0,
        'g/gal' : 1.0 * 1000 / 42.0,
        'mg/gal' : 1.0 * 1000 * 1000 / 42.0,
        'metricTon/gal' : 1.0 / 1000.0 / 42.0,
        'lb/gal' : 2.2046226218 / 42.0,

        'kg/m3' : 1.0 / 0.158987294928,
        'g/m3' : 1.0 * 1000 / 0.158987294928,
        'mg/m3' : 1.0 * 1000 * 1000 / 0.158987294928,
        'metricTon/m3' : 1.0 / 1000.0 / 0.158987294928,
        'lb/m3' : 2.2046226218 / 0.158987294928,

        'kg/lt' : 1.0 / 158.987294928,
        'g/lt' : 1.0 * 1000 / 158.987294928,
        'mg/lt' : 1.0 * 1000 * 1000 / 158.987294928,
        'metricTon/lt' : 1.0 / 1000.0 / 158.987294928,
        'lb/lt' : 2.2046226218 / 158.987294928,

        'kg/ml' : 1.0 / 158987.294928,
        'g/ml' : 1.0 * 1000 / 158987.294928,
        'mg/ml' : 1.0 * 1000 * 1000 / 158987.294928,
        'metricTon/ml' : 1.0 / 1000.0 / 158987.294928,
        'lb/ml' : 2.2046226218 / 158987.294928,
    }

    def __init__(self, value, unit):

        if unit not in Density.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Density.Units.list()}")
        
        super(Density, self).__init__(value=value, unit=unit)


from flask import request
from datetime import datetime
from flask_restx import Namespace, Resource
from .models.users import signup_parser, login_parser
from .... import PyAutomation, TIMEZONE, _TIMEZONE
from ....extensions.api import api
from ....extensions import _api as Api
from ....modules.users.users import Users as CVTUsers
from ....dbmodels.users import Users

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S"
ns = Namespace('Users', description='User Management and Authentication')
app = PyAutomation()
users = CVTUsers()

@ns.route('/')
class UsersCollection(Resource):

    @api.doc(security='apikey', description="Retrieves a list of all registered users.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self):
        """
        Get all users.

        Retrieves a list of all users currently registered in the system.
        """

        return users.serialize(), 200

@ns.route('/signup')
class SignUpResource(Resource):
    
    @Api.validate_reqparser(reqparser=signup_parser)
    @api.doc(security='apikey', description="Registers a new user.")
    @api.response(200, "User created successfully")
    @api.response(400, "User creation failed")
    @ns.expect(signup_parser)
    def post(self):
        """
        User signup.

        Registers a new user with the provided credentials and role.
        """
        args = signup_parser.parse_args()
        user, message = app.signup(**args)
        
        if user:

            return user.serialize(), 200
        
        return message, 400


@ns.route('/login')
class LoginResource(Resource):

    @Api.validate_reqparser(reqparser=login_parser)
    @api.doc(security='apikey', description="Authenticates a user and returns an API token.")
    @api.response(200, "Login successful")
    @api.response(403, "Invalid credentials")
    @ns.expect(login_parser)
    def post(self):
        """
        User login.

        Authenticates a user using username/email and password. Returns an API key/token.
        """
        args = login_parser.parse_args()
        user, message = app.login(**args)

        if user:

            return {
                "apiKey": user.token,
                "role": user.role.name,
                "role_level": user.role.level,
                "datetime": datetime.now(TIMEZONE).strftime(DATETIME_FORMAT),
                "timezone": _TIMEZONE
                }, 200

        return message, 403
    

@ns.route('/credentials_are_valid')
class VerifyCredentialsResource(Resource):
    
    @api.doc(security='apikey', description="Verifies if the provided credentials are valid without logging in.")
    @api.response(200, "Success (True/False)")
    @Api.token_required(auth=True)
    @Api.validate_reqparser(reqparser=login_parser)
    @ns.expect(login_parser)
    def post(self):
        """
        Verify credentials.

        Checks if the provided username/password combination is valid.
        """
        args = login_parser.parse_args()
        credentials_valid, _ = users.verify_credentials(**args)
        return credentials_valid, 200
    
@ns.route('/<username>')
@api.param('username', 'The username')
class UserResource(Resource):
    
    @api.doc(security='apikey', description="Retrieves information about a specific user.")
    @api.response(200, "Success")
    @api.response(400, "User not found")
    @Api.token_required(auth=True)
    def get(self, username):
        """
        Get user info.

        Retrieves detailed information about a specific user by username.
        """
        
        user = users.get_by_username(username=username)

        if user:

            return user.serialize(), 200

        return f"{username} is not a valid username", 400


@ns.route('/logout')
class LogoutResource(Resource):

    @api.doc(security='apikey', description="Logs out the current user and invalidates the token.")
    @api.response(200, "Logout successful")
    @Api.token_required(auth=True)
    def post(self):
        """
        User logout.

        Invalidates the current session token.
        """
        if 'X-API-KEY' in request.headers:
                            
            token = request.headers['X-API-KEY']

        elif 'Authorization' in request.headers:
            
            token = request.headers['Authorization'].split('Token ')[-1]
        
        _, message = Users.logout(token=token)

        return message, 200


import dash
from ...pages.components.opcua import OPCUAComponents
from ...opcua.subscription import SubHandler
from ...models import StringType
from ...state_machine import Node, ua
from ...utils import find_differences_between_lists_opcua_server


subscription_handler = SubHandler()
opcua_components = OPCUAComponents()


def init_callback(app:dash.Dash):

    def create_opcua_server_table(opcua_server_machine):

        attrs = list()        
        for attr in dir(opcua_server_machine):
            if hasattr(opcua_server_machine, attr):
                node = getattr(opcua_server_machine, attr)
                if isinstance(node, Node):

                    node_class = node.get_node_class()
                    if node_class == ua.NodeClass.Variable:
                        
                        display_name = node.get_attribute(ua.AttributeIds.DisplayName).Value.Value.Text
                        # Obtén el nodo padre
                        parent_node = node.get_parent()

                        # Obtén el nombre de la carpeta padre (nodo padre)
                        parent_name = parent_node.get_browse_name().Name
                        access_level = node.get_access_level()
                        # Verificar los niveles de acceso
                        write_only = ua.AccessLevel.CurrentWrite in access_level and ua.AccessLevel.CurrentRead not in access_level
                        read_write = ua.AccessLevel.CurrentRead in access_level and ua.AccessLevel.CurrentWrite in access_level
                        access_type = "Read"
                        if write_only:

                            access_type = "Write"
                        
                        elif read_write:

                            access_type = "ReadWrite"

                        attrs.append({
                            "name": f"{parent_name}.{display_name}",
                            "namespace": node.nodeid.to_string(),
                            "access_type": access_type
                        })

                        properties = node.get_properties()
                        for prop in properties:
                            prop_name = prop.get_display_name().Text
                            # prop_value = prop.get_value()

                            access_level = prop.get_access_level()
                            # Verificar los niveles de acceso
                            write_only = ua.AccessLevel.CurrentWrite in access_level and ua.AccessLevel.CurrentRead not in access_level
                            read_write = ua.AccessLevel.CurrentRead in access_level and ua.AccessLevel.CurrentWrite in access_level
                            access_type = "Read"
                            if write_only:

                                access_type = "Write"
                            
                            elif read_write:

                                access_type = "ReadWrite"

                            attrs.append({
                                "name": f"{parent_name}.{display_name}.{prop_name}",
                                "namespace": prop.nodeid.to_string(),
                                "access_type": access_type
                            })
        return attrs

    @app.callback(
        dash.Output("opcua_server_datatable", "data", allow_duplicate=True),
        dash.Input('opcua_server', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        attrs = list()

        if pathname=="/opcua-server":
            opcua_server_machine = app.automation.get_machine(name=StringType("OPCUAServer"))
            attrs = create_opcua_server_table(opcua_server_machine=opcua_server_machine)

        return attrs
    
    @app.callback(
        dash.Input('opcua_server_datatable', 'data_timestamp'),
        dash.State('opcua_server_datatable', 'data_previous'),
        dash.State('opcua_server_datatable', 'data'),
        )
    def update_read_only(timestamp, previous, current):
        message = None
        attr_not_clearable = ("name", "namespace")
        if timestamp:

            if previous and current: # UPDATE TAG DEFINITION
                
                to_updates = find_differences_between_lists_opcua_server(previous, current)
                node_to_update = to_updates[0]
                node_name = node_to_update.pop("name")
                node_to_update.pop("namespace")
                for attr in attr_not_clearable:
                    if attr in node_to_update:
                        if not node_to_update[attr]:
                            message = f"You can not empty {attr} attribute"
                
                if message:
                    dash.set_props("modal-update-opcua-server-body", {"children": message})
                    dash.set_props("modal-update-opcua-server-centered", {'is_open': True})
                    return
                message = f"Do you want to update node {node_name} Access Type to {node_to_update['access_type']}?"
                # OPEN MODAL TO CONFIRM CHANGES
                dash.set_props("modal-update-opcua-server-body", {"children": message})
                dash.set_props("modal-update-opcua-server-centered", {'is_open': True})

    @app.callback(
        [
            dash.Output("modal-update-opcua-server-centered", "is_open"), 
            dash.Output('opcua_server_datatable', 'data'), 
            dash.Output('opcua_server_datatable', 'data_timestamp'),
            dash.Output("update-opcua-server-yes", "n_clicks"),
            dash.Output("update-opcua-server-no", "n_clicks")
        ],
        [dash.Input("update-opcua-server-yes", "n_clicks"), dash.Input("update-opcua-server-no", "n_clicks")],
        [
            dash.State('opcua_server_datatable', 'data_timestamp'),
            dash.State("modal-update-opcua-server-centered", "is_open"),
            dash.State('opcua_server_datatable', 'data_previous'),
            dash.State('opcua_server_datatable', 'data')
        ]
    )
    def toggle_modal_update_read_only(yes_n, no_n, timestamp, is_open, previous, current):
        r"""
        Documentation here
        """
        from ...opcua.subscription import SubHandlerServer

        handler = SubHandlerServer()
        opcua_server_machine = app.automation.get_machine(name=StringType("OPCUAServer"))
        attrs = create_opcua_server_table(opcua_server_machine=opcua_server_machine)

        if yes_n:
            
            if timestamp:
                        
                if previous and current: # UPDATE TAG DEFINITION
                    to_updates = find_differences_between_lists_opcua_server(previous, current)
                    node_to_update = to_updates[0]
                    namespace = node_to_update.pop("namespace")
                    access_type = node_to_update.pop("access_type")
                    # Code for update read_only attribute
                    opcua_server_machine = app.automation.get_machine(name=StringType("OPCUAServer"))
                    opcua_server_attrs = dir(opcua_server_machine)
                    node = False

                    for i, item in enumerate(opcua_server_attrs):
                    
                        if hasattr(opcua_server_machine, item):
                            node = getattr(opcua_server_machine, item)
                            if isinstance(node, Node):

                                node_class = node.get_node_class()

                                if node_class == ua.NodeClass.Variable:

                                    if node.nodeid.to_string()==namespace:
                                        
                                        break 

                                    else:
                                        props = node.get_properties()
                                        flag = False
                                        for node in props:
                                            
                                            if node.nodeid.to_string()==namespace:
                                                
                                                flag = True
                                                break

                                        if flag:

                                            break

                    if node:

                        opcua_server_obj = app.automation.get_opcua_server_record_by_namespace(namespace=namespace)
                        if opcua_server_obj:
                            app.automation.update_opcua_server_access_type(namespace=namespace, access_type=access_type)
                        else:
                            app.automation.create_opcua_server_record(name=node_to_update["name"], namespace=namespace, access_type=access_type)
                        access_type = access_type.lower()
                        # Limpiar todos los bits de acceso primero
                        node.unset_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.CurrentRead)
                        node.unset_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.CurrentWrite)
                        node.unset_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.CurrentRead)
                        node.unset_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.CurrentWrite)

                        subscriptions = handler.subscriptions
                        # Unsubscribe

                        if namespace in subscriptions:

                            _sub = subscriptions.pop(namespace)
                            _sub.delete()
                        
                        if access_type == "write":
                            # Solo escritura: deshabilitamos la lectura y habilitamos la escritura
                            node.set_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.CurrentWrite)
                            node.set_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.CurrentWrite)
                            sub = opcua_server_machine.server.create_subscription(100, handler)
                            sub.subscribe_data_change(node)
                            handler.subscriptions[namespace] = sub
                        elif access_type == "read":
                            # Solo lectura: habilitamos la lectura y deshabilitamos la escritura
                            node.set_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.CurrentRead)
                            node.set_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.CurrentRead)
                        elif access_type == "readwrite":
                            # Lectura y escritura: habilitamos ambos
                            node.set_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.CurrentRead)
                            node.set_attr_bit(ua.AttributeIds.AccessLevel, ua.AccessLevel.CurrentWrite)
                            node.set_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.CurrentRead)
                            node.set_attr_bit(ua.AttributeIds.UserAccessLevel, ua.AccessLevel.CurrentWrite)
                            sub = opcua_server_machine.server.create_subscription(100, handler)
                            sub.subscribe_data_change(node)
                            handler.subscriptions[namespace] = sub

                    attrs = create_opcua_server_table(opcua_server_machine=opcua_server_machine)

                return not is_open, attrs, None, 0, 0

        elif no_n:
            
            return not is_open, attrs, None, 0, 0

        else:

            return is_open, attrs, None, 0, 0



import logging
from .observer import Observer
from ..variables import VARIABLES
from ..alarms.states import ACTIONS

def log_detailed(e, message):
    logger = logging.getLogger("pyautomation")
    logger.error(message)
    logger.error(e, exc_info=True)

def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def find_differences_between_lists(prev_list, curr_list):
    r"""
    Documentation here
    """
    
    differences = []

    for prev_dict, curr_dict in zip(prev_list, curr_list):
        diff = {'id': prev_dict['id']}
        for key in prev_dict:
            if prev_dict[key] != curr_dict[key]:
                diff[key] = curr_dict[key]
        if len(diff) > 1:  # Only add if there are differences other than 'id'
            differences.append(diff)
    
    return differences

def find_differences_between_lists_opcua_server(prev_list, curr_list):
    r"""
    Documentation here
    """
    
    differences = []

    for prev_dict, curr_dict in zip(prev_list, curr_list):
        diff = {'name': prev_dict['name'], 'namespace': prev_dict['namespace']}

        for key in prev_dict:
            if prev_dict[key] != curr_dict[key]:
                diff[key] = curr_dict[key]
        if len(diff) > 2:  # Only add if there are differences other than 'id'
            differences.append(diff)
    
    return differences

def find_keys_values_by_unit(d, unit):
    r"""
    Documentation here
    """
    result = []
    
    for _, sub_dict in d.items():
        
        if unit in sub_dict.values():
            
            for k, v in sub_dict.items():
                
                result.append({'label': k, 'value': v})

    return result

def generate_dropdown_conditional():
    r"""
    Documentation here
    """
    from .. import PyAutomation
    app = PyAutomation()
    data = VARIABLES
    dropdown_conditional = []
    # dropdown conditional for unit column in table
    for variable, sub_dict in data.items():

        for unit in sub_dict.values():

            options = find_keys_values_by_unit(data, unit)
            
            dropdown_conditional.append({
                'if': {'column_id': 'unit', 'filter_query': f'{{variable}} eq "{variable}"'},
                'options': options,
                'clearable': False
            })

            dropdown_conditional.append({
                'if': {'column_id': 'display_unit', 'filter_query': f'{{variable}} eq "{variable}"'},
                'options': options,
                'clearable': False
            })

    # dropdown conditional for node_namespace column in table
    dropdown_conditional.append({
        'if': {'column_id': 'node_namespace', 'filter_query': f'{{opcua_address}} eq ""'},
        'options': []
    })
    for client_name, info in app.get_opcua_clients().items():
        
        tree = app.get_opcua_tree(client_name=client_name)
        options = list()
        for node in tree[0]["Objects"][0]["children"]:
            options.append({
                "label": node["title"],
                "value": node["key"]
            })
            server = info['server_url']
            dropdown_conditional.append({
                'if': {'column_id': 'node_namespace', 'filter_query': f'{{opcua_address}} eq "{server}"'},
                'options': options
            })

    return dropdown_conditional

def get_nodes_info(selected_files:list):
    r"""
    Documentation here
    """
    to_get_node_values = dict()
    for file in selected_files:
        
        if file:

            info = file[0].split("/")
            client_name = info[0]
            namespace = info[1]
            
            if client_name in to_get_node_values:

                to_get_node_values[client_name].append(namespace)

            else:

                to_get_node_values[client_name] = [namespace]

    return to_get_node_values

def get_data_to_update_into_opcua_table(app, to_get_node_values:dict):
    r"""
    Documentation here
    """
    data = list()
    for client_name, namespaces in to_get_node_values.items():
        
        infos = app.automation.get_node_attributes(client_name=client_name, namespaces=namespaces)

        if infos:
        
            for info in infos:
                _info = info[0]
                namespace = _info["Namespace"]
                data.append(
                    {
                        "server": client_name,
                        "namespace": namespace,
                        "data_type": _info["DataType"],
                        "display_name": _info["DisplayName"],
                        "value": _info["Value"],
                        "source_timestamp": _info["DataValue"].SourceTimestamp,
                        "status_code": _info["DataValue"].StatusCode.name
                    }
                )
    
    return data


from .api import Api
from .cors import Cors

_api = Api()
_cors = Cors()

def init_app(app):
    """
    Application extensions initialization.
    """
    extensions = ( _api, _cors)

    for extension in extensions:
        
        extension.init_app(app)


from peewee import CharField, ForeignKeyField
from ..dbmodels.core import BaseModel

class AccessType(BaseModel):
    r"""
    Database model for OPC UA Node Access Types (Read, Write, ReadWrite).
    """
    
    name = CharField(unique=True)

    @classmethod
    def create(cls, name:str="Read")-> dict:
        r"""
        Creates a new Access Type.

        **Parameters:**

        * **name** (str): Access type name.

        **Returns:**

        * **AccessType**: The created or existing record.
        """
        if name.lower()=="read" or name.lower()=="write" or name.lower()=="readwrite":

            access_type_obj = cls.read_by_name(name=name)
            
            if not access_type_obj:
                query = cls(name=name)
                query.save()  
                return query
            
        return cls.read_by_name(name=name)

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
        Retrieves an Access Type by name.
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Checks if an Access Type name exists.
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serializes the record.
        """

        return {
            "id": self.id,
            "name": self.name
        }


class OPCUAServer(BaseModel):
    r"""
    Database model for OPC UA Server Node configurations.
    """
    
    name = CharField(unique=True)
    namespace = CharField(unique=True)
    access_type = ForeignKeyField(AccessType, null=True)

    @classmethod
    def create(cls, name:str, namespace:str, access_type:str)-> dict:
        r"""
        Creates a new OPC UA Server Node record.

        **Parameters:**

        * **name** (str): Node name.
        * **namespace** (str): Node namespace/ID.
        * **access_type** (str): Access level.

        **Returns:**

        * **OPCUAServer**: The created record.
        """
        if not cls.name_exist(name=name):
            
            if not cls.namespace_exist(namespace=namespace):
                
                if AccessType.name_exist(name=access_type):
                    access_type_obj = AccessType.read_by_name(name=access_type)
                else:
                    access_type_obj = AccessType.create(name=access_type)
                
                if access_type_obj:

                    query = cls(
                        name=name,
                        namespace=namespace,
                        access_type=access_type_obj
                        )
                    query.save()

                    return query

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
        Retrieves a node by name.
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query
        
        return None
    
    @classmethod
    def read_by_namespace(cls, namespace:str)->bool:
        r"""
        Retrieves a node by namespace.
        """
        query = cls.get_or_none(namespace=namespace)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Checks if a node name exists.
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False
    
    @classmethod
    def namespace_exist(cls, namespace:str)->bool:
        r"""
        Checks if a node namespace exists.
        """
        query = cls.get_or_none(namespace=namespace)
        
        if query is not None:

            return True
        
        return False
    
    @classmethod
    def update_access_type(cls, namespace:str, access_type:str)-> dict:
        r"""
        Updates the access type of an existing node.

        **Parameters:**

        * **namespace** (str): The node namespace.
        * **access_type** (str): The new access type.
        """
        obj = cls.get_or_none(namespace=namespace)
        
        if obj:
            
            if AccessType.name_exist(name=access_type):
                access_type_obj = AccessType.read_by_name(name=access_type)
            else:
                access_type_obj = AccessType.create(name=access_type)

            if access_type_obj:
                
                query = cls.update(access_type=access_type_obj).where(cls.id == obj.id)
                query.execute()
                return query

    def serialize(self)-> dict:
        r"""
        Serializes the node record.
        """

        return {
            "id": self.id,
            "name": self.name,
            "access_type": self.access_type.serialize()
        }


# -*- coding: utf-8 -*-
"""automation/managers/alarms.py

This module implements the Alarm Manager, which is responsible for managing alarm definitions,
handling alarm events, and interacting with the Current Value Table (CVT) and Database.
"""
from datetime import datetime
import queue
from ..singleton import Singleton
from ..tags import CVTEngine, TagObserver
from ..alarms import AlarmState, Alarm
from ..dbmodels.alarms import AlarmSummary
from ..modules.users.users import User
from ..models import FloatType, StringType
from ..utils.decorators import set_event, logging_error_handler
from flask_socketio import SocketIO


class AlarmManager(Singleton):
    r"""
    Singleton class that manages all alarms in the system.

    It handles the creation, update, deletion, and retrieval of alarms.
    It also validates trigger conditions and manages communication with the frontend via SocketIO.
    """

    def __init__(self):

        self._alarms:dict[Alarm] = dict()
        self._tag_queue = queue.Queue()
        self.tag_engine = CVTEngine()

    def get_queue(self)->queue.Queue:
        r"""
        Retrieves the internal tag queue used for observer notifications.

        **Returns:**

        * **queue.Queue**: The queue instance.
        """
        return self._tag_queue

    @logging_error_handler
    def append_alarm(
            self,
            name:str,
            tag:str,
            type:str="BOOL",
            trigger_value:bool|float=True,
            description:str="",
            identifier:str=None,
            state:str="Normal",
            timestamp:str=None,
            ack_timestamp:str=None,
            user:User=None,
            reload:bool=False,
            sio:SocketIO|None=None
        )->tuple[Alarm, str]:
        r"""
        Creates and registers a new alarm in the manager.

        **Parameters:**

        * **name** (str): Alarm name.
        * **tag** (str): Associated Tag name.
        * **type** (str): Alarm type (BOOL, HH, H, L, LL).
        * **trigger_value** (bool|float): Value that triggers the alarm.
        * **description** (str, optional): Alarm description.
        * **identifier** (str, optional): Unique ID.
        * **state** (str, optional): Initial state.
        * **timestamp** (str, optional): Last trigger timestamp.
        * **ack_timestamp** (str, optional): Last acknowledgment timestamp.
        * **user** (User, optional): User creating the alarm.
        * **reload** (bool, optional): If reloading from DB.
        * **sio** (SocketIO, optional): SocketIO instance for real-time updates.

        **Returns:**

        * **tuple[Alarm, str]**: The created Alarm object and a status message.
        """
        # Check alarm name duplicated
        alarm = self.get_alarm_by_name(name)
        if alarm:

            return alarm, f"Alarm {name} is already defined"

        # Check if alarm is associated to same tag with same alarm type
        trigger_value_message = self.__check_trigger_values(name=name, tag=tag, type=type, trigger_value=trigger_value)
        if trigger_value_message:

            return None, trigger_value_message

        if timestamp:

            timestamp = datetime.strptime(timestamp, self.tag_engine.DATETIME_FORMAT)

        if ack_timestamp:

            ack_timestamp = datetime.strptime(ack_timestamp, self.tag_engine.DATETIME_FORMAT)

        alarm = Alarm(
            name=name,
            tag=self.tag_engine.get_tag_by_name(name=tag),
            description=description,
            alarm_type=StringType(type),
            alarm_setpoint=FloatType(trigger_value),
            identifier=identifier,
            state=state,
            timestamp=timestamp,
            ack_timestamp=ack_timestamp,
            user=user,
            reload=reload
        )
        alarm.set_socketio(sio=sio)
        self._alarms[alarm.identifier] = alarm

        return alarm, f"Alarm creation successful"

    @logging_error_handler
    def put(
            self,
            id:str,
            name:str=None,
            tag:str=None,
            description:str=None,
            alarm_type:str=None,
            trigger_value:float=None,
            user:User=None
            )->tuple[Alarm, str]:
        r"""
        Updates an existing alarm configuration.

        **Parameters:**

        * **id** (str): Alarm identifier.
        * **name** (str, optional): New name.
        * **tag** (str, optional): New tag name.
        * **description** (str, optional): New description.
        * **alarm_type** (str, optional): New alarm type.
        * **trigger_value** (float, optional): New trigger value.
        * **user** (User, optional): User performing the update.

        **Returns:**

        * **tuple[Alarm, str]**: The updated Alarm object and a status message.
        """
        alarm = self.get_alarm(id=id)
        if name:

            if self.get_alarm_by_name(name=name):

                return f"Alarm {name} is already defined"

        # Check if alarm is associated to same tag with same alarm type
        if not tag:
            tag = alarm.tag
        if not alarm_type:
            alarm_type = alarm.alarm_setpoint.type
        if not trigger_value:
            trigger_value = FloatType(alarm.alarm_setpoint.value)

        trigger_value_message = self.__check_trigger_values(
            name=alarm.name,
            tag=tag,
            type=alarm_type,
            trigger_value=trigger_value
            )
        if trigger_value_message:

            return None, trigger_value_message

        alarm, message = alarm.put(
            user=user,
            name=name,
            tag=tag,
            description=description,
            alarm_type=alarm_type,
            trigger_value=trigger_value
            )
        self._alarms[id] = alarm

    @logging_error_handler
    @set_event(message=f"Deleted", classification="Alarm", priority=3, criticity=5)
    def delete_alarm(self, id:str, user:User=None):
        r"""
        Removes an alarm from the manager and takes it out of service.

        **Parameters:**

        * **id** (str): Alarm ID.
        * **user** (User, optional): User performing the deletion.
        """
        if id in self._alarms:

            alarm = self._alarms.pop(id)
            alarm.remove_from_service(user=user)

        return alarm, f"Alarm: {alarm.name} - Tag: {alarm.tag}"

    @logging_error_handler
    def get_alarm(self, id:str)->Alarm:
        r"""
        Retrieves an alarm by its ID.

        **Parameters:**

        * **id** (str): Alarm ID.

        **Returns:**

        * **Alarm**: The alarm object if found.
        """

        if id in self._alarms:

            return self._alarms[id]

    @logging_error_handler
    def get_alarm_by_name(self, name:str)->Alarm:
        r"""
        Retrieves an alarm by its name.

        **Parameters:**

        * **name** (str): Alarm name.

        **Returns:**

        * **Alarm**: The alarm object if found.
        """
        for id, alarm in self._alarms.items():

            if name == alarm.name:

                return self._alarms[str(id)]

    @logging_error_handler
    def get_alarms_by_tag(self, tag:str)->dict:
        r"""
        Retrieves all alarms associated with a specific tag (by name).

        **Parameters:**

        * **tag** (str): Tag name.

        **Returns:**

        * **dict**: A dictionary of {id: Alarm} objects.
        """
        alarms = dict()
        for id, alarm in self._alarms.items():

            if tag == alarm.tag:

                alarms[id] = alarm

        return alarms

    @logging_error_handler
    def get_alarm_by_tag(self, tag:str)->list[Alarm]:
        r"""
        Retrieves a list of alarms associated with a specific tag.

        **Parameters:**

        * **tag** (str): Tag name.

        **Returns:**

        * **list[Alarm]**: List of Alarm objects.
        """
        alarms = list()
        for _, alarm in self._alarms.items():

            if tag == alarm.tag:

                alarms.append(alarm)

        return alarms

    @logging_error_handler
    def get_alarms(self)->dict:
        r"""
        Retrieves all registered alarms.

        **Returns:**

        * **dict**: Dictionary of all Alarm objects.
        """
        return self._alarms

    @logging_error_handler
    def get_lasts_active_alarms(self, lasts:int=None)->list:
        r"""
        Retrieves the most recent active alarms.

        **Parameters:**

        * **lasts** (int, optional): Number of alarms to retrieve.

        **Returns:**

        * **list**: List of serialized active alarms sorted by timestamp.
        """
        original_list = [alarm.serialize() for _, alarm in self.get_alarms().items()]
        filtered_list = [elem for elem in original_list if elem['state']['alarm_status'].lower()=="active"]
        sorted_list = sorted(filtered_list, key=lambda x: x['timestamp'] if x['timestamp'] else '')
        if lasts:

            if len(sorted_list)>lasts:

                sorted_list = sorted_list[0:lasts]

        return sorted_list

    @logging_error_handler
    def serialize(self)->list:
        r"""
        Serializes all alarms managed by this instance.

        **Returns:**

        * **list**: List of serialized alarm dictionaries.
        """

        return [alarm.serialize() for _, alarm in self._alarms.items()]

    @logging_error_handler
    def get_tag_alarms(self)->list:
        r"""
        Retrieves a list of Tags that have alarms associated with them.

        **Returns:**

        * **list**: List of Tag objects.
        """
        result = [_alarm.tag_alarm for id, _alarm in self.get_alarms().items()]

        return result

    @logging_error_handler
    def tags(self)->list:
        r"""
        Retrieves a unique list of Tag names bound to alarms.

        **Returns:**

        * **list**: List of Tag names.
        """
        result = set([_alarm.tag for id, _alarm in self.get_alarms().items()])

        return list(result)

    @logging_error_handler
    def __check_trigger_values(self, name:str, tag:str, type:str, trigger_value:float)->None|str:
        r"""
        Validates trigger values to prevent logical conflicts (e.g., Low limit > High limit).

        **Parameters:**

        * **name** (str): Name of the new/updated alarm.
        * **tag** (str): Tag name.
        * **type** (str): Alarm type.
        * **trigger_value** (float): Trigger threshold.

        **Returns:**

        * **None|str**: None if valid, or an error message string if invalid.
        """
        alarms = self.get_alarm_by_tag(tag=tag)

        if alarms:

            for alarm in alarms:

                if alarm.name!=name:

                    if type==alarm.alarm_setpoint.type.value:

                        return f"Alarm Type {type} and alarm's tag {tag} duplicated"

                    if type=="LOW-LOW":

                        if trigger_value>=alarm.alarm_setpoint.value:

                            return f"Conflict definition with {alarm.name} in trigger value {trigger_value}>={alarm.alarm_setpoint.value}"

                    if type=="LOW":

                        if alarm.alarm_setpoint.type.value=="LOW-LOW":

                            if trigger_value<=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}>={alarm.alarm_setpoint.value}"

                        else:

                            if trigger_value>=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}>={alarm.alarm_setpoint.value}"

                    if type=="HIGH":

                        if alarm.alarm_setpoint.type.value=="HIGH-HIGH":

                            if trigger_value>=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}<={alarm.alarm_setpoint.value}"

                        else:

                            if trigger_value<=alarm.alarm_setpoint.value:

                                return f"Conflict definition with {alarm.name} in trigger value {trigger_value}<={alarm.alarm_setpoint.value}"

                    if type=="HIGH-HIGH":

                        if trigger_value<=alarm.alarm_setpoint.value:

                            return f"Conflict definition with {alarm.name} in trigger value {trigger_value}<={alarm.alarm_setpoint.value}"

    @logging_error_handler
    def filter_by(self, **fields):
        r"""
        Filters historical alarms via the database model.

        **Parameters:**

        * **fields**: Filtering criteria (name, state, timestamp, etc.).

        **Returns:**

        * **tuple**: (Result data, HTTP status code 200).
        """

        return AlarmSummary.filter_by(**fields), 200

    @logging_error_handler
    def get_lasts(self, lasts:int=10):
        r"""
        Retrieves the last N alarm summary records.

        **Parameters:**

        * **lasts** (int): Number of records.

        **Returns:**

        * **tuple**: (List of records, HTTP status code 200).
        """

        return AlarmSummary.read_lasts(lasts=lasts), 200

    @logging_error_handler
    def summary(self)->dict:
        r"""
        Generates a summary of the current alarm manager state.

        **Returns:**

        * **dict**: Summary including total alarms, alarm names, and associated tags.
        """
        result = dict()
        alarms = [_alarm.name for id, _alarm in self.get_alarms().items()]
        result["length"] = len(alarms)
        result["alarms"] = alarms
        result["alarm_tags"] = self.get_tag_alarms()
        result["tags"] = self.tags()

        return result

    @logging_error_handler
    def attach(self, alarm_name:str):
        r"""
        Attaches a tag observer to a specific alarm's tag.

        **Parameters:**

        * **alarm_name** (str): Name of the alarm.
        """
        def attach_observer(entity):

            _tag = entity.tag

            observer = TagObserver(self._tag_queue)
            self.tag_engine.attach(name=_tag, observer=observer)

        alarm = self.get_alarm_by_name(name=alarm_name)
        attach_observer(alarm)

    @logging_error_handler
    def execute(self, tag_name:str):
        r"""
        Evaluates alarm conditions for a given tag based on its current value.
        
        Also handles auto-unshelving of alarms if their shelved duration has expired.

        **Parameters:**

        * **tag_name** (str): Name of the tag to evaluate.
        """
        value = self.tag_engine.get_value_by_name(tag_name=tag_name)['value']

        for _, _alarm in self._alarms.items():

            if _alarm.state == AlarmState.SHLVD:

                _now = datetime.now()

                if _alarm._shelved_until:

                    if _now >= _alarm._shelved_until:

                        _alarm.unshelve()
                        continue

                    continue

                continue

            if tag_name==_alarm.tag:

                _alarm.update(value)



import unittest, os
from datetime import datetime
from .. import PyAutomation
from ..alarms import Alarm, AlarmState


class TestCore(unittest.TestCase):

    def setUp(self) -> None:
        file_path = os.path.join(".", "db", "test.db")
        if os.path.exists(file_path):
            os.remove(file_path)
        self.app = PyAutomation()
        self.app.run(debug=True, test=True, create_tables=True)
        return super().setUp()

    def tearDown(self) -> None:
        self.app.safe_stop()
        return super().tearDown()
    
    def test_tags(self):
        
        # CREATE TAGS
        _tag1 = {
            "name": "P1",
            "unit": "Pa",
            "variable": "Pressure"
        }
        tag1, _ = self.app.create_tag(**_tag1)
        with self.subTest("Test tag in cvt"):
            
            tag_in_cvt = self.app.cvt.get_tag_by_name(name=_tag1['name'])
            self.assertEqual(tag1, tag_in_cvt)

        # GET TAGS
        _tag2 = {
            "name": "T1",
            "unit": "C",
            "variable": "Temperature"
        }
        tag2, _ = self.app.create_tag(**_tag2)

        # SET TAG VALUES
        timestamp = datetime.now()
        value = 35
        self.app.cvt.set_value(id=tag2.id, value=value, timestamp=timestamp)
        with self.subTest("Test Value in CVT"):
            
            self.assertEqual(self.app.cvt.get_value(id=tag2.id), value)

        # UPDATE TAGS
        name = "TT"
        updated_tag, _ = self.app.update_tag(id=tag2.id, name=name)
        with self.subTest("Test update tag name"):

            self.assertEqual(name, updated_tag.name)

        with self.subTest("Test update tag name from DB"):

            updated_tag = self.app.logger_engine.get_tag_by_name(name=name)
            self.assertEqual(name, updated_tag.name)

        unit = "K"
        updated_tag, _ = self.app.update_tag(id=tag2.id, unit=unit)
        with self.subTest("Test update tag unit"):

            self.assertEqual(unit, updated_tag.unit)

        with self.subTest("Test update tag unit from DB"):

            updated_tag = self.app.logger_engine.get_tag_by_name(name=tag2.name)
            self.assertEqual(unit, updated_tag.unit.unit)

        # DELETE TAG
        self.app.delete_tag(id=tag2.id)
        with self.subTest("Test delete tag"):

            self.assertIsNone(self.app.get_tag_by_name(name=tag2.name))

        with self.subTest("Test delete tag from DB"):
            tag = self.app.logger_engine.get_tag_by_name(name=tag2.name)
            self.assertFalse(tag.active)

        # DELETE TAG BY NAME
        self.app.delete_tag_by_name(name=tag1.name)
        with self.subTest("Test delete tag by name"):

            self.assertIsNone(self.app.get_tag_by_name(name=tag1.name))

        with self.subTest("Test delete tag by name from DB"):
            tag = self.app.logger_engine.get_tag_by_name(name=tag1.name)
            self.assertFalse(tag.active)

    def test_alarms(self):
        r"""
        Documentation here
        """

        tag_payload = {
            "name": "T2",
            "unit": "C",
            "variable": "Temperature"
        }
        tag, _ = self.app.create_tag(**tag_payload)

        alarm_LL_payload = {
            "name": "alarm_LL",
            "tag": tag.name,
            "alarm_type": "LOW-LOW",
            "trigger_value": 10.0,
        }

        # CREATE ALARM
        alarm_LL, _ = self.app.create_alarm(**alarm_LL_payload)
        with self.subTest("Test create alarm instance"):

            self.assertIsInstance(alarm_LL, Alarm)

        with self.subTest("Test create alarm in alarm manager"):

            self.assertEqual(alarm_LL, self.app.alarm_manager.get_alarm_by_name(name=alarm_LL.name))

        alarm_L_payload = {
            "name": "alarm_L",
            "tag": tag.name,
            "alarm_type": "LOW",
            "trigger_value": 20.0,
        }
        alarm_L, _ = self.app.create_alarm(**alarm_L_payload)
        with self.subTest("Test create alarm instance"):

            self.assertIsInstance(alarm_L, Alarm)
        
        with self.subTest("Test create alarm in alarm manager"):

            self.assertEqual(alarm_L, self.app.alarm_manager.get_alarm_by_name(name=alarm_L.name))

        alarm_H_payload = {
            "name": "alarm_H",
            "tag": tag.name,
            "alarm_type": "HIGH",
            "trigger_value": 30.0,
        }
        alarm_H, _ = self.app.create_alarm(**alarm_H_payload)
        with self.subTest("Test create alarm instance"):

            self.assertIsInstance(alarm_H, Alarm)

        with self.subTest("Test create alarm in alarm manager"):

            self.assertEqual(alarm_H, self.app.alarm_manager.get_alarm_by_name(name=alarm_H.name))

        alarm_HH_payload = {
            "name": "alarm_HH",
            "tag": tag.name,
            "alarm_type": "HIGH-HIGH",
            "trigger_value": 40.0,
        }
        alarm_HH, _ = self.app.create_alarm(**alarm_HH_payload)
        with self.subTest("Test create alarm instance"):

            self.assertIsInstance(alarm_HH, Alarm)

        with self.subTest("Test create alarm in alarm manager"):

            self.assertEqual(alarm_HH, self.app.alarm_manager.get_alarm_by_name(name=alarm_HH.name))

        # UPDATE ALARM DEFINITION
        self.app.update_alarm(id=alarm_HH.identifier, trigger_value=50)
        with self.subTest("Test update alarm in Alarm Manager"):
            
            self.assertEqual(alarm_HH.alarm_setpoint.value, 50)

        with self.subTest("Test update alarm in DB"):
            alarm = self.app.alarms_engine.get_alarm_by_name(name=alarm_HH.name)
            self.assertEqual(alarm.trigger_value, 50)

        # TRIGGER ALARMS
        timestamp = datetime.now()
        self.app.cvt.set_value(id=tag.id, value=35, timestamp=timestamp)
        with self.subTest("Test Trigger HIGH Alarm"):
            
            self.assertEqual(alarm_H.state.alarm_status, "Active")

        with self.subTest("Test check UNACK alarm state"):
            
            self.assertEqual(alarm_H.state, AlarmState.UNACK)

        with self.subTest("Test aknowledge alarm"):
            alarm_H.acknowledge()
            self.assertEqual(alarm_H.state, AlarmState.ACKED)

        with self.subTest("Test not Trigger HIGH-HIGH Alarm"):

            self.assertEqual(alarm_HH.state.alarm_status, "Not Active")

        self.app.cvt.set_value(id=tag.id, value=0, timestamp=timestamp)
        with self.subTest("Test Trigger LOW Alarm"):
            
            self.assertEqual(alarm_L.state.alarm_status, "Active")

        with self.subTest("Test Trigger LOW-LOW Alarm"):
            
            self.assertEqual(alarm_LL.state.alarm_status, "Active")

        with self.subTest("Test check UNACK alarm LL state"):
            
            self.assertEqual(alarm_LL.state, AlarmState.UNACK)

        with self.subTest("Test check UNACK alarm L state"):
            
            self.assertEqual(alarm_L.state, AlarmState.UNACK)

        self.app.cvt.set_value(id=tag.id, value=15, timestamp=timestamp)
        with self.subTest("Test check UNACK alarm LL state"):
            
            self.assertEqual(alarm_LL.state, AlarmState.RTNUN)

        with self.subTest("Test aknowledge LL alarm"):
            alarm_LL.acknowledge()
            self.assertEqual(alarm_LL.state, AlarmState.NORM)

        with self.subTest("Test aknowledge L alarm"):
            alarm_L.acknowledge()
            self.assertEqual(alarm_L.state, AlarmState.ACKED)


        # Boolean Alarm
        tag_payload = {
            "name": "Bool1",
            "unit": "adim",
            "variable": "Adimentional"
        }
        tag, _ = self.app.create_tag(**tag_payload)
        alarm_B_payload = {
            "name": "alarm_B",
            "tag": tag.name,
            "alarm_type": "BOOL",
            "trigger_value": True,
        }
        alarm_B, _ = self.app.create_alarm(**alarm_B_payload)

        with self.subTest("Test Not Trigger BOOL Alarm"):
            
            self.assertEqual(alarm_B.state.alarm_status, "Not Active")

        self.app.cvt.set_value(id=tag.id, value=True, timestamp=timestamp)
        with self.subTest("Test Trigger BOOL Alarm"):
            
            self.assertEqual(alarm_B.state.alarm_status, "Active")

        # DELETE TAG
        self.app.delete_alarm(id=alarm_LL.identifier)
        self.app.delete_alarm(id=alarm_L.identifier)
        self.app.delete_alarm(id=alarm_H.identifier)
        self.app.delete_alarm(id=alarm_HH.identifier)
        self.app.delete_tag(id=tag.id)


# -*- coding: utf-8 -*-
"""automation/logger/users.py

This module implements the Users Logger, responsible for persisting user accounts,
role definitions, and logging user authentication events.
"""
from ..modules.users.users import User
from ..dbmodels import Users, Roles
from .core import BaseEngine, BaseLogger
from ..utils.decorators import db_rollback


class UsersLogger(BaseLogger):
    r"""
    Logger class specialized for User Management persistence.
    """

    def __init__(self):

        super(UsersLogger, self).__init__()

    @db_rollback
    def set_role(self, name:str, level:int, identifier:str):
        r"""
        Creates a new user role in the database.

        **Parameters:**

        * **name** (str): Role name (e.g., "admin").
        * **level** (int): Permission level (e.g., 100).
        * **identifier** (str): Unique role ID.
        """
        return Roles.create(
            name=name,
            level=level,
            identifier=identifier
        )

    @db_rollback
    def set_user(
            self, 
            user:User
        ):
        r"""
        Creates a new user in the database.

        **Parameters:**

        * **user** (User): User object containing name, password, email, roles.
        """
        return Users.create(
            user=user
        )
    
    @db_rollback
    def login(
            self, 
            password:str,
            username:str="",
            email:str=""
        ):
        r"""
        Authenticates a user against the database.

        **Parameters:**

        * **password** (str): Plain text password.
        * **username** (str, optional): Username.
        * **email** (str, optional): Email.

        **Returns:**

        * **tuple**: (User object, Message string).
        """
        return Users.login(
            password=password,
            username=username,
            email=email
        )
    
class UsersLoggerEngine(BaseEngine):
    r"""
    Thread-safe Engine for the UsersLogger.
    """

    def __init__(self):

        super(UsersLoggerEngine, self).__init__()
        self.logger = UsersLogger()

    def set_role(self, name:str, level:int, identifier:str):
        r"""
        Thread-safe role creation.
        """
        _query = dict()
        _query["action"] = "set_role"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["level"] = level
        _query["parameters"]["identifier"] = identifier
        
        return self.query(_query)
    
    def set_user(self, user:User):
        r"""
        Thread-safe user creation.
        """
        _query = dict()
        _query["action"] = "set_user"
        _query["parameters"] = dict()
        _query["parameters"]["user"] = user
        
        return self.query(_query)
    
    def login(self,password:str, username:str="", email:str=""):
        r"""
        Thread-safe login.
        """
        _query = dict()
        _query["action"] = "login"
        _query["parameters"] = dict()
        _query["parameters"]["password"] = password
        _query["parameters"]["username"] = username
        _query["parameters"]["email"] = email
        
        return self.query(_query)



from ..utils.decorators import decorator
import numpy as np

class KalmanFilter:
    r"""
    Implements a standard Kalman Filter for single-variable estimation.

    The Kalman Filter estimates the state of a system from a series of incomplete and noisy measurements.
    """
    def __init__(
            self, 
            x, 
            A:np.array=np.array([[1]]), 
            B:np.array=np.array([[0]]), 
            H:np.array=np.array([[1]]), 
            P:np.array=np.array([[1]]), 
            Q:np.array=np.array([[1e-5]]), 
            R:np.array=np.array([[0.5]])
            ):
        r"""
        Initializes the Kalman Filter.

        **Parameters:**

        * **x**: Initial state estimate.
        * **A**: State transition model.
        * **B**: Control input model.
        * **H**: Observation model.
        * **P**: Estimate covariance.
        * **Q**: Process noise covariance.
        * **R**: Measurement noise covariance.
        """
        self.A = A
        self.B = B
        self.H = H
        self.Q = Q
        self.R = R
        self.P = P
        self.x = x
        self.previous_innov = None                      # Para guardar la innovación anterior

    def predict(self, u=0):
        r"""
        Predicts the next state of the system.

        **Parameters:**

        * **u**: Control input vector (optional).
        """
        self.x = np.dot(self.A, self.x) + np.dot(self.B, u)
        self.P = np.dot(np.dot(self.A, self.P), self.A.T) + self.Q

    def update(self, z, threshold:float=100, r_value:float=0.5):
        r"""
        Updates the state estimate with a new measurement.

        Includes adaptive logic to adjust measurement noise (R) based on innovation variance.

        **Parameters:**

        * **z**: Measurement value.
        * **threshold** (float): Threshold for innovation variance to adjust R.
        * **r_value** (float): Default R value.
        """
        
        innov = z - np.dot(self.H, self.x)  # Innovación
        if self.previous_innov is not None:
            innov_var = np.std([self.previous_innov, innov])  # Varianza de las innovaciones
            self.R = 0.0 if innov_var > threshold else r_value  # Ajustar R
        K = np.dot(np.dot(self.P, self.H.T), np.linalg.inv(np.dot(np.dot(self.H, self.P), self.H.T) + self.R))
        self.x = self.x + np.dot(K, innov)
        self.P = self.P - np.dot(np.dot(K, self.H), self.P)
        self.previous_innov = innov


class GaussianFilter:
    r"""
    A wrapper for the Kalman Filter designed for simple scalar value filtering.
    
    It maintains the filter state between calls.
    """

    def __init__(self):

        self.kf = None

    def __call__(self, value:float, threshold:float=100, r_value:float=0.5):
        r"""
        Applies the filter to a new value.

        **Parameters:**

        * **value** (float): The noisy input value.
        * **threshold** (float): Threshold for adaptive filtering.
        * **r_value** (float): Measurement noise parameter.

        **Returns:**

        * **float**: The filtered value.
        """
        
        if self.kf is None:

            self.kf = KalmanFilter(value)

        self.kf.predict()
        self.kf.update(value, threshold=threshold, r_value=r_value)
        filtered_value = self.kf.x[0][0]
        return filtered_value



from flask_restx import reqparse

login_parser = reqparse.RequestParser(bundle_errors=True)
login_parser.add_argument("username", type=str, required=False, help='Username')
login_parser.add_argument("email", type=str, required=False, help='User email')
login_parser.add_argument("password", type=str, required=True, help='User password')

signup_parser = reqparse.RequestParser(bundle_errors=True)
signup_parser.add_argument("username", type=str, required=True, help='Username')
signup_parser.add_argument("role_name", type=str, required=True, help='Role Name')
signup_parser.add_argument("email", type=str, required=True, help="Email address")
signup_parser.add_argument("password", type=str, required=True, help="Password")
signup_parser.add_argument("name", type=str, required=False, help="User's name")
signup_parser.add_argument("lastname", type=str, required=False, help="User's lastname")


# -*- coding: utf-8 -*-
"""automation/logger/alarms.py

This module implements the Alarms Logger, responsible for persisting alarm definitions,
alarm status history, and summaries to the database.
"""
from datetime import datetime
from ..dbmodels import Alarms, AlarmSummary, AlarmTypes, AlarmStates
from .core import BaseEngine, BaseLogger
from ..alarms.trigger import TriggerType
from ..alarms.states import AlarmState
from ..utils.decorators import db_rollback


class AlarmsLogger(BaseLogger):
    r"""
    Logger class specialized for Alarm Management.

    It handles CRUD operations for Alarm definitions and logging of Alarm events (activations, acknowledgments, etc.).
    """

    def __init__(self):

        super(AlarmsLogger, self).__init__()

    @db_rollback
    def create_tables(self, tables):
        r"""
        Creates alarm-related tables and initializes default alarm types and states.

        **Parameters:**

        * **tables** (list): List of database models.
        """
        if not self.check_connectivity():
            
            return
        
        self._db.create_tables(tables, safe=True)
        self.__init_default_alarms_schema()

    @db_rollback
    def __init_default_alarms_schema(self):
        r"""
        Initializes default Alarm Types (High, Low, etc.) and Alarm States (Active, Ack, etc.) in the DB.
        """
        ## Alarm Types
        for alarm_type in TriggerType:

            AlarmTypes.create(name=alarm_type.value)

        ## Alarm States
        for alarm_state in AlarmState._states:
            name = alarm_state.state
            mnemonic = alarm_state.mnemonic
            condition = alarm_state.process_condition
            status = alarm_state.alarm_status
            AlarmStates.create(name=name, mnemonic=mnemonic, condition=condition, status=status)

    @db_rollback
    def create(
            self,
            id:str,
            name:str,
            tag:str,
            trigger_type:str,
            trigger_value:float,
            description:str):
        r"""
        Creates a new Alarm definition in the database.

        **Parameters:**

        * **id** (str): Alarm unique identifier.
        * **name** (str): Alarm name.
        * **tag** (str): Associated tag name.
        * **trigger_type** (str): Type of trigger (e.g., "HIGH", "LOW").
        * **trigger_value** (float): The threshold value.
        * **description** (str): Description of the alarm.
        """
        if not self.check_connectivity():
            
            return 

        Alarms.create(
            identifier=id,
            name=name,
            tag=tag,
            trigger_type=trigger_type,
            trigger_value=trigger_value,
            description=description
        )

    @db_rollback
    def get_alarms(self):
        r"""
        Retrieves all configured alarms.

        **Returns:**

        * **list**: List of Alarm model instances.
        """
        if not self.check_connectivity():
            
            return list()
        
        alarms = Alarms.read_all()

        if alarms:

            return alarms
    
    @db_rollback
    def get_alarm_by_name(self, name:str)->Alarms|None:
        r"""
        Retrieves a specific alarm by name.

        **Parameters:**

        * **name** (str): Alarm name.

        **Returns:**

        * **Alarms**: The alarm model instance or None.
        """
        if not self.check_connectivity():
            
            return None
        
        return Alarms.read_by_name(name=name)

    @db_rollback        
    def get_lasts(self, lasts:int=10):
        r"""
        Retrieves the last N entries from the Alarm Summary (history).

        **Parameters:**

        * **lasts** (int): Number of entries to retrieve.

        **Returns:**

        * **list**: List of AlarmSummary entries.
        """
        if not self.is_history_logged:

            return list()
        
        if not self.check_connectivity():
            
            return list()
        
        return AlarmSummary.read_lasts(lasts=lasts)
    
    @db_rollback
    def filter_alarm_summary_by(
            self,
            states:list[str]=None,
            names:list[str]=None,
            tags:list[str]=None,
            greater_than_timestamp:datetime=None,
            less_than_timestamp:datetime=None,
            timezone:str="UTC",
            page:int=1,
            limit:int=20
        ):
        r"""
        Filters alarm history based on various criteria.

        **Parameters:**

        * **states** (list[str]): Filter by alarm states.
        * **names** (list[str]): Filter by alarm names.
        * **tags** (list[str]): Filter by tag names.
        * **greater_than_timestamp** (datetime): Start time.
        * **less_than_timestamp** (datetime): End time.
        * **timezone** (str): Timezone.
        * **page** (int): Pagination page.
        * **limit** (int): Entries per page.

        **Returns:**

        * **list**: Filtered list of alarm summaries.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
        
        return AlarmSummary.filter_by(
            states=states,
            names=names,
            tags=tags,
            greater_than_timestamp=greater_than_timestamp,
            less_than_timestamp=less_than_timestamp,
            timezone=timezone,
            page=page,
            limit=limit
        )
    
    @db_rollback
    def put(
        self,
        id:str,
        name:str=None,
        tag:str=None,
        description:str=None,
        alarm_type:str=None,
        trigger_value:str=None,
        state:str=None
        ):
        r"""
        Updates an existing alarm definition.

        **Parameters:**

        * **id** (str): Alarm ID.
        * **name** (str, optional): New name.
        * **tag** (str, optional): New tag.
        * **description** (str, optional): New description.
        * **alarm_type** (str, optional): New alarm type.
        * **trigger_value** (str, optional): New trigger value.
        * **state** (str, optional): New state.
        """
        if not self.check_connectivity():
            
            return None
        
        fields = dict()
        alarm = Alarms.read_by_identifier(identifier=id)
        if alarm:
            if name:
                fields["name"] = name
            if tag:
                fields["tag"] = tag
            if description:
                fields["description"] = description
            if alarm_type:
                alarm_type = AlarmTypes.read_by_name(name=alarm_type)
                fields["trigger_type"] = alarm_type
            if trigger_value:
                fields["trigger_value"] = trigger_value
            if state:
                alarm_state = AlarmStates.get_or_none(name=state)
                fields["state"] = alarm_state
            query = Alarms.put(
                id=alarm.id,
                **fields
            )

            return query

    @db_rollback
    def delete(self, id:str):
        r"""
        Logically deletes an alarm (sets it to "Out Of Service").

        **Parameters:**

        * **id** (str): Alarm ID.
        """
        if not self.check_connectivity():
            
            return None
        
        alarm_state = AlarmStates.get_or_none(name="Out Of Service")
        alarm = Alarms.read_by_identifier(identifier=id)
        Alarms.put(
            id=alarm.id,
            state=alarm_state
        )

    @db_rollback
    def create_record_on_alarm_summary(self, name:str, state:str, timestamp:datetime, ack_timestamp:datetime=None):
        r"""
        Creates a new entry in the Alarm Summary (history log).

        **Parameters:**

        * **name** (str): Alarm name.
        * **state** (str): Alarm state.
        * **timestamp** (datetime): Timestamp of the event.
        * **ack_timestamp** (datetime, optional): Acknowledgment timestamp.
        """
        if not self.is_history_logged:

            return None
        
        if self.check_connectivity():
            
            AlarmSummary.create(name=name, state=state, timestamp=timestamp, ack_timestamp=ack_timestamp)

    @db_rollback
    def put_record_on_alarm_summary(self, name:str, state:str=None, ack_timestamp:datetime=None):
        r"""
        Updates the latest record in the Alarm Summary for a given alarm.

        **Parameters:**

        * **name** (str): Alarm name.
        * **state** (str, optional): New state.
        * **ack_timestamp** (datetime, optional): Acknowledgment timestamp.
        """
        if not self.check_connectivity():
            
            return None
        
        if not self.is_history_logged:

            return None
        
        fields = dict()
        alarm = AlarmSummary.read_by_name(name=name)

        if alarm:

            if ack_timestamp:
                fields["ack_time"] = ack_timestamp
            if state:
                alarm_state = AlarmStates.get_or_none(name=state)
                fields["state"] = alarm_state

            if fields:
                query = AlarmSummary.put(
                    id=alarm.id,
                    **fields
                )

                return query

    @db_rollback
    def get_alarm_summary(self):
        r"""
        Retrieves the entire alarm summary.

        **Returns:**

        * **list**: List of all AlarmSummary records.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
        
        return AlarmSummary.read_all()
    
    
class AlarmsLoggerEngine(BaseEngine):
    r"""
    Thread-safe Engine for the AlarmsLogger.
    """

    def __init__(self):

        super(AlarmsLoggerEngine, self).__init__()
        self.logger = AlarmsLogger()

    def create(
        self,
        id:str,
        name:str,
        tag:str,
        trigger_type:str,
        trigger_value:float,
        description:str
        ):
        r"""
        Thread-safe alarm creation.
        """
        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["name"] = name
        _query["parameters"]["tag"] = tag
        _query["parameters"]["trigger_type"] = trigger_type
        _query["parameters"]["trigger_value"] = trigger_value
        _query["parameters"]["description"] = description
        
        return self.query(_query)
    
    def get_lasts(
        self,
        lasts:int=1
        ):
        r"""
        Thread-safe retrieval of last alarm events.
        """
        _query = dict()
        _query["action"] = "get_lasts"
        _query["parameters"] = dict()
        _query["parameters"]["lasts"] = lasts
        
        return self.query(_query)
    
    def get_alarms(self):
        r"""
        Thread-safe retrieval of all alarms.
        """
        _query = dict()
        _query["action"] = "get_alarms"
        _query["parameters"] = dict()
        
        return self.query(_query)
    
    def get_alarm_by_name(self, name:str):
        r"""
        Thread-safe retrieval of an alarm by name.
        """
        _query = dict()
        _query["action"] = "get_alarm_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        
        return self.query(_query)
    
    def filter_alarm_summary_by(
        self,
        names:list[str]=None,
        states:list[int]=None,
        tags:list[int]=None,
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):
        r"""
        Thread-safe filtering of alarm summary.
        """
        _query = dict()
        _query["action"] = "filter_alarm_summary_by"
        _query["parameters"] = dict()
        _query["parameters"]["names"] = names
        _query["parameters"]["states"] = states
        _query["parameters"]["tags"] = tags
        _query["parameters"]["greater_than_timestamp"] = greater_than_timestamp
        _query["parameters"]["less_than_timestamp"] = less_than_timestamp
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["page"] = page
        _query["parameters"]["limit"] = limit
        
        return self.query(_query)
    
    def create_record_on_alarm_summary(self, name:str, state:str, timestamp:datetime, ack_timestamp:datetime=None):
        r"""
        Thread-safe creation of alarm history record.
        """
        _query = dict()
        _query["action"] = "create_record_on_alarm_summary"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["state"] = state
        _query["parameters"]["timestamp"] = timestamp
        _query["parameters"]["ack_timestamp"] = ack_timestamp
        
        return self.query(_query)
    
    def put_record_on_alarm_summary(
        self,
        name:str,
        state:str=None,
        ack_timestamp:datetime=None
        ):
        r"""
        Thread-safe update of alarm history record.
        """
        _query = dict()
        _query["action"] = "put_record_on_alarm_summary"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["state"] = state
        _query["parameters"]["ack_timestamp"] = ack_timestamp

        return self.query(_query)

    def put(
        self,
        id:str,
        name:str=None,
        tag:str=None,
        description:str=None,
        alarm_type:str=None,
        trigger_value:str=None,
        state:str=None
        ):
        r"""
        Thread-safe alarm update.
        """
        _query = dict()
        _query["action"] = "put"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["name"] = name
        _query["parameters"]["tag"] = tag
        _query["parameters"]["description"] = description
        _query["parameters"]["alarm_type"] = alarm_type
        _query["parameters"]["trigger_value"] = trigger_value
        _query["parameters"]["state"] = state

        return self.query(_query)

    def delete(self, id:str):
        r"""
        Thread-safe alarm deletion.
        """
        _query = dict()
        _query["action"] = "delete"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.query(_query)

    def get_alarm_summary(self):
        r"""
        Thread-safe retrieval of entire alarm summary.
        """
        _query = dict()
        _query["action"] = "get_alarm_summary"
        _query["parameters"] = dict()
        
        return self.query(_query)

    def create_tables(self, tables):
        r"""
        Thread-safe table creation.
        """
        self.logger.create_tables(tables)



# -*- coding: utf-8 -*-
"""automation/logger/machines.py

This module implements the Machines Logger, responsible for persisting State Machine
configurations and their bindings to Tags.
"""
from ..dbmodels import Machines, TagsMachines, Tags
from .core import BaseEngine, BaseLogger
from ..utils.decorators import db_rollback
from ..models import IntegerType, StringType, FloatType
from ..tags.tag import Tag


class MachinesLogger(BaseLogger):
    r"""
    Logger class specialized for State Machine persistence.
    """

    def __init__(self):

        super(MachinesLogger, self).__init__()

    @db_rollback
    def create(
            self,
            identifier:str,
            name:str,
            interval:int,
            description:str,
            classification:str,
            buffer_size:int,
            buffer_roll_type:str,
            criticity:int,
            priority:int,
            on_delay:int=None,
            threshold:float=None
            ):
        r"""
        Creates a new State Machine definition in the database.

        **Parameters:**

        * **identifier** (str): Unique ID.
        * **name** (str): Machine name.
        * **interval** (int): Execution interval.
        * **description** (str): Description.
        * **classification** (str): Type of machine.
        * **buffer_size** (int): Size of internal buffer.
        * **buffer_roll_type** (str): Roll type (e.g., "fifo").
        * **criticity** (int): Criticality level.
        * **priority** (int): Priority level.
        * **on_delay** (int, optional): Delay before starting.
        * **threshold** (float, optional): Operational threshold.
        """
        if not self.check_connectivity():

            return None
        
        if hasattr(threshold, "value"):
            
            threshold = threshold.value
       
        Machines.create(
            identifier=identifier,
            name=name,
            interval=interval,
            description=description,
            classification=classification,
            buffer_size=buffer_size,
            buffer_roll_type=buffer_roll_type,
            criticity=criticity,
            priority=priority,
            on_delay=on_delay,
            threshold=threshold
        )

    @db_rollback
    def put(
        self,
        name:StringType,
        machine_interval:IntegerType=None,
        description:StringType=None,
        classification:StringType=None,
        buffer_size:IntegerType=None,
        buffer_roll_type:StringType=None,
        criticity:IntegerType=None,
        priority:IntegerType=None,
        on_delay:IntegerType=None,
        threshold:FloatType=None
        ):
        r"""
        Updates an existing State Machine definition.

        **Parameters:**

        * **name** (StringType): Machine name.
        * **machine_interval** (IntegerType, optional): New interval.
        * **description** (StringType, optional): New description.
        * **classification** (StringType, optional): New classification.
        * **buffer_size** (IntegerType, optional): New buffer size.
        * **buffer_roll_type** (StringType, optional): New roll type.
        * **criticity** (IntegerType, optional): New criticality.
        * **priority** (IntegerType, optional): New priority.
        * **on_delay** (IntegerType, optional): New on delay.
        * **threshold** (FloatType, optional): New threshold.
        """

        if not self.check_connectivity():
            
            return None

        fields = dict()
        machine = Machines.read_by_name(name=name.value)
        if machine_interval:
            
            fields["interval"] = machine_interval.value
        if description:
            fields["description"] = description.value
        if classification:
            fields["classification"] = classification.value
        if buffer_size:
            fields["buffer_size"] = buffer_size.value
        if buffer_roll_type:
            fields["buffer_roll_type"] = buffer_roll_type.value
        if criticity:
            fields["criticity"] = criticity.value
        if priority:
            fields["priority"] = priority.value
        if on_delay:
            fields["on_delay"] = on_delay.value
        if threshold:
            if hasattr(threshold.value, "value"):
                threshold.value = threshold.value.value
            fields["threshold"] = threshold.value
            
        query = Machines.put(
            id=machine.id,
            **fields
        )

        return query
    
    @db_rollback
    def read_all(self):
        r"""
        Retrieves all machine definitions.
        """
        if not self.check_connectivity():

            return list()
        
        return Machines.read_all()
    
    @db_rollback
    def read_config(self):
        r"""
        Retrieves machine configuration for the scheduler.
        """
        if not self.check_connectivity():
            
            return None

        return Machines.read_config()
    
    @db_rollback
    def bind_tag(self, tag:Tag, machine, default_tag_name:str=None):
        r"""
        Binds a Tag to a State Machine in the database.

        **Parameters:**

        * **tag** (Tag): The Tag object.
        * **machine** (StateMachine): The Machine object.
        * **default_tag_name** (str, optional): Default tag alias within the machine.
        """
        if not self.check_connectivity():

            return None
            
        TagsMachines.create(tag_name=tag.name, machine_name=machine.name.value, default_tag_name=default_tag_name)

    @db_rollback
    def unbind_tag(self, tag:Tag, machine):
        r"""
        Unbinds a Tag from a State Machine.
        """
        if not self.check_connectivity():

            return None

        tag_from_db = Tags.get_or_none(name=tag.name)
        machine_from_db= Machines.get_or_none(name=machine.name.value)
        tags_machine = TagsMachines.get((TagsMachines.tag == tag_from_db) & (TagsMachines.machine == machine_from_db))
        tags_machine.delete_instance()    

class MachinesLoggerEngine(BaseEngine):
    r"""
    Thread-safe Engine for the MachinesLogger.
    """

    def __init__(self):

        super(MachinesLoggerEngine, self).__init__()
        self.logger = MachinesLogger()

    def create(
        self,
        identifier:str,
        name:str,
        interval:int,
        description:str,
        classification:str,
        buffer_size:int,
        buffer_roll_type:str,
        criticity:int,
        priority:int,
        on_delay:int=None,
        threshold:float=None
        ):
        r"""
        Thread-safe machine creation.
        """
        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["identifier"] = identifier
        _query["parameters"]["name"] = name
        _query["parameters"]["interval"] = interval
        _query["parameters"]["classification"] = classification
        _query["parameters"]["description"] = description
        _query["parameters"]["buffer_size"] = buffer_size
        _query["parameters"]["buffer_roll_type"] = buffer_roll_type
        _query["parameters"]["criticity"] = criticity
        _query["parameters"]["priority"] = priority
        _query["parameters"]["on_delay"] = on_delay
        _query["parameters"]["threshold"] = threshold
        
        return self.query(_query)
    
    def put(
        self,
        name:StringType,
        machine_interval:IntegerType=None,
        description:StringType=None,
        classification:StringType=None,
        buffer_size:IntegerType=None,
        buffer_roll_type:StringType=None,
        criticity:IntegerType=None,
        priority:IntegerType=None,
        on_delay:IntegerType=None,
        threshold:FloatType=None
        ):
        r"""
        Thread-safe machine update.
        """
        _query = dict()
        _query["action"] = "put"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["machine_interval"] = machine_interval
        _query["parameters"]["classification"] = classification
        _query["parameters"]["description"] = description
        _query["parameters"]["buffer_size"] = buffer_size
        _query["parameters"]["buffer_roll_type"] = buffer_roll_type
        _query["parameters"]["criticity"] = criticity
        _query["parameters"]["priority"] = priority
        _query["parameters"]["on_delay"] = on_delay
        _query["parameters"]["threshold"] = threshold

        return self.query(_query)

    def read_all(self):
        r"""
        Thread-safe read all machines.
        """
        _query = dict()
        _query["action"] = "read_all"
        _query["parameters"] = dict()
        return self.query(_query)
    
    def read_config(self):
        r"""
        Thread-safe read config.
        """
        _query = dict()
        _query["action"] = "read_config"
        _query["parameters"] = dict()
        return self.query(_query)
    
    def bind_tag(self, tag:Tag, machine, default_tag_name:str=None):
        r"""
        Thread-safe bind tag.
        """
        _query = dict()
        _query["action"] = "bind_tag"
        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        _query["parameters"]["machine"] = machine
        _query["parameters"]["default_tag_name"] = default_tag_name
        return self.query(_query)
    
    def unbind_tag(self, tag:Tag, machine):
        r"""
        Thread-safe unbind tag.
        """
        _query = dict()
        _query["action"] = "unbind_tag"
        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        _query["parameters"]["machine"] = machine
        return self.query(_query)



import dash
import dash_bootstrap_components as dbc
from automation.pages.components import Components
from automation.pages.components.database import DatabaseComponents

dash.register_page(__name__)

layout = dbc.Container(
    [
        dbc.Breadcrumb(
            items=[
                {"label": "Home", "href": "/"},
                {"label": "Database", "active": True},
            ],
        ),
        DatabaseComponents.create_db_config_form(),
        Components.modal_error(
            title="Database Connection",
            modal_id="modal-db-connection",
            button_close_id="close-model-db-connection",
            body_id="modal-body-db-connection"
        )
    ],
    fluid=False,
    className="my-3"
)
